####################################################################################
#                                                                                  #
# Copyright (c) 2006 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#                                                                                  #
#  This file is globally imported into Picalo.  See picalo/__init__.py.            # 
#  When you write "from picalo import *", these functions all get imported.        #
#                                                                                  #
####################################################################################

import os, sys, time, types, codecs, os.path, tempfile, inspect
from picalo import check_valid_table, show_progress, clear_progress, make_unicode, useProgress, use_progress_indicators
from Column import Column, _ColumnLoader
from Record import Record
from Expression import PicaloExpression
from Error import error
import ZODB.DB
import persistent
from zc.blist import BList
from persistent.dict import PersistentDict
from persistent.list import PersistentList

# global functions (defined in this file)
__all__ = [
  'Table',
]


#################################
###   A Picalo table

class Table(persistent.Persistent):
  '''The primary data structure used in Picalo modules.  This is the first object a user
     should learn about since it's used everywhere. Tables are the input into most
     functions and are also the return records of most functions. 
     
     See the Picalo manual for more information on tables.   
     
     The columns (field definitions) are required to be named and to have types.
     Fields are specified as a sequence of (name, type, format).  The name
     can be any string starting with a letter and then any combination
     of letters and numbers.  Column names must be unique.
     The type must be a type object, such as int, float, unicode, DateTime, etc.  
     The format is a mask using special characters; see the set_format() method
     for documentation on the different masks.
     
     When specifying field definitions, the type and format are optional.
     Therefore, a field definition can be described as any of the following:
     
     - name      - If a simple string is passed into the field, it is assumed
                   to be the column name.  The field type defaults to string
                   and no format is used.
     - ( name )  - A string passed in within a sequence is also assumed to be
                   the column name.  The field type defaults ot string and no
                   format is used.
     - ( name, type ) - The column name and type.  No format is used.
     - ( name, type, format ) - Explicitly setting the name, type and format. 
     - ( name, type, format, static_expression, active_expression) - Explicitly setting the 
         name, type, format, expression, and/or active expression.  Since it doesn't make 
         sense for a column to be both a static expression and active expression, one
         of these should be None.  See examples below.
     
     The initial data for the table can be specified as a sequence of sequences,
     or a grid of data.  See below for examples.

     Example 1:
       >>> # creates a table with two columns, defaulting to a string type and no format. 
       >>> data = project.create_table([ 
       ...  id,
       ...  name,
       ... ])
       >>> data.structure().view()
       +--------+------------------+------------+--------+
       | Column |       Type       | Expression | Format |
       +--------+------------------+------------+--------+
       | id     | <type 'unicode'> |    <N>     |  <N>   |
       | name   | <type 'unicode'> |    <N>     |  <N>   |
       +--------+------------------+------------+--------+         

     Example 2:
       >>> # creates a table with two columns
       >>> data = project.create_table([ 
       ...  ( 'id',   int ),
       ...  ( 'name', unicode),
       ... ])
       >>> data.structure().view()
       +--------+------------------+------------+--------+
       | Column |       Type       | Expression | Format |
       +--------+------------------+------------+--------+
       | id     | <type 'int'>     |    <N>     |  <N>   |
       | name   | <type 'unicode'> |    <N>     |  <N>   |
       +--------+------------------+------------+--------+         
       
     Example 3:
       >>> # creates a table with two columns, an initial format on the ID field, and initial data
       >>> data = Table([ 
       ...   ( 'id',    number,    '#.#0' ),
       ...   ( 'name',  unicode        ),
       ... ],[
       ...   [ 1.253, 'Danny'    ],
       ...   [ 2,     'Vijay'    ],
       ...   [ 3,     'Dongsong' ],
       ...   [ 4,     'Sally'    ],
       ... ])
       >>> data.structure().view()
       +--------+------------------+------------+--------+
       | Column |       Type       | Expression | Format |
       +--------+------------------+------------+--------+
       | id     | <type 'number'>  |    <N>     |  #.#0  |
       | name   | <type 'unicode'> |    <N>     |  <N>   |
       +--------+------------------+------------+--------+         
     
     Example 4:
       >>> # creates a table with two columns (using shortcut notation)
       >>> # since types are not specified, both columns are typed as unicode
       >>> data = Table(['id', 'name'])
       >>> data.structure().view()
       +--------+------------------+------------+--------+
       | Column |       Type       | Expression | Format |
       +--------+------------------+------------+--------+
       | id     | <type 'unicode'> |    <N>     |  <N>   |
       | name   | <type 'unicode'> |    <N>     |  <N>   |
       +--------+------------------+------------+--------+
       
     Example 5:
       >>> # creates a table with three columns (using another shortcut notation)
       >>> data = Table(3)
       >>> data.structure().view()
       +--------+------------------+------------+--------+
       | Column |       Type       | Expression | Format |
       +--------+------------------+------------+--------+
       | col000 | <type 'unicode'> |    <N>     |  <N>   |
       | col001 | <type 'unicode'> |    <N>     |  <N>   |
       | col002 | <type 'unicode'> |    <N>     |  <N>   |
       +--------+------------------+------------+--------+
       
     Example 6:
       >>> # creates a table with an active calculated column. This column will update
       >>> # automatically as other values in the table change
       >>> data = Table([
       >>>   ( 'c1', int, None),
       >>>   ( 'c4', int, None, None, 'c1*3'),
       >>> ])
       >>> data.append(1)
       >>> data.append(2)
       >>> data.view()
       +----+----+
       | c1 | c4 |
       +----+----+
       |  1 |  3 |
       |  2 |  6 |
       +----+----+
       >>> data[0].c1 = 5
       >>> data.view()
       +----+----+
       | c1 | c2 |
       +----+----+
       |  5 | 15 |
       |  2 |  6 |
       +----+----+

     Example 7:
       >>> # creates a table with a static calculated column. Picalo will assign the results
       >>> # of the given expression to the field and will never again update them.
       >>> # This is different than example 6 because the values are static.
       >>> data = Table([
       >>>   ( 'c1', int, None),
       >>>   ( 'c2', int, None, 'c1*3'),
       >>> ], [
       >>>   [ 1 ],
       >>>   [ 2 ],
       >>> ])
       >>> data.view()        
       +----+----+
       | c1 | c4 |
       +----+----+
       |  1 |  3 |
       |  2 |  6 |
       +----+----+
       >>> data.view()
       >>> data[0].c1 = 5
       +----+----+
       | c1 | c2 |
       +----+----+
       |  5 |  3 |
       |  2 |  6 |
       +----+----+
  '''
  TRANSIENT_VARIABLE_NAMES = [ '_listeners', '_indexes', 'changed', 'filterindex', 'filterexpression' ]
  
  def __init__(self, columns=3, data=[]):
    '''Creates a Picalo Table.  Users should not call this method directly.
       Instead, call Project.create_table() or Project.open_table().
    '''
    # check the parameter types
    assert isinstance(columns, (types.IntType, types.LongType, types.ListType, types.TupleType, PersistentList)), 'The columns parameter must be list of (name, type, format).'
    assert isinstance(data, (Table, types.ListType, types.TupleType)), 'The initial data of a table must be a list of lists, such as [ [1, "a"], [2, "b"] ]'
    # set up the class variables
    self._data = BList()  # a very efficient, ZODB-aware list
    self._listeners = []   # listeners to be notified when changes occur to the table (for integration into GUI)
    self._indexes = {}     # indices used by Simple.select_by_record and other functions
    self.readonly = False
    self.changed = False   # whether the data in the table has changed
    self.columns_map_names = PersistentDict()  # index that maps column names to real column numbers
    self.columns_map_virtual = PersistentDict()  # index that maps virtual column numbers to real column numbers
    self.columns = PersistentList()      # the list of Column objects that represent the columns
    self.num_columns = 0          # the number of active columns we have
    self.filterindex = None    
    self.filterexpression = None
    # set up the columns
    if isinstance(columns, (types.IntType, types.LongType)):  # switch to a list if columns is an integer
      columns = [ 'col%03i' % i for i in range(columns) ]
    for i, col in enumerate(columns):
      if isinstance(col, (Column, _ColumnLoader)): 
        self.columns.append(Column(self, i, col.name, col.column_type, col.expression, col.format))
      elif isinstance(col, (types.ListType, types.TupleType)): 
        assert isinstance(col[0], types.StringTypes), 'Column name must be a string in: ' + str(col)
        col = list(col)  # ensure we have a list 
        while len(col) < 5:
          col.append(None)
        self.columns.append(Column(self, i, col[0], col[1], col[4], col[2]))
      elif isinstance(col, types.StringTypes): 
        self.columns.append(Column(self, i, col, unicode))
      else:
        raise AssertionError, 'Invalid (name, type, format) specification: ' + str(col)
    self._calculate_columns_map()

    # add any initial data to the table
    if data:
      self.extend(data)

    # now that we have data, calculate a static expression if we've been asked to
    # the only way this would do anything is if initial data was placed in the table
    for i, col in enumerate(columns):
      if isinstance(col, (types.ListType, types.TupleType)) and not isinstance(col, (Column, _ColumnLoader)) and len(col) >= 4 and col[3]:
        self.columns[i].static_expression = col[3]
        self.replace_column_values(i, col[3])
      
      
  def __setstate__(self, state):
    '''Restores the object from ZODB and initializes transient variables'''
    persistent.Persistent.__setstate__(self, state)
    # restore the transient variables to their defaults
    self._listeners = []
    self._indexes = {}
    self.changed = False
    self.filterindex = None
    self.filterexpression = None
    # we have to access *any* method of Record so ZODB loads the record and column information from disk
    if len(self) > 0:
      len(self[0])
    
    
  def __getstate__(self):
    '''Returns the state of the object for ZODB serialization.  First removes transient variables'''
    c = persistent.Persistent.__getstate__(self)
    for varname in self.TRANSIENT_VARIABLE_NAMES: # remove any variables that are transient
      del c[varname]
    return c
        
      
  def _add_listener(self, listener):
    '''Adds a listener to the table.  Will be notified when data changes occur.
       The listener should be a callable/function of form callback(table).  For
       efficiency reasons and because we don't need it right now,
       the col and row is not reported.'''
    self._remove_listener(listener)  # just in case it's already here
    self._listeners.append(listener)
    
    
  def _remove_listener(self, listener):
    '''Removes a listener from the table.'''
    try:
      while True:  # continue until no more of this listener in our list
        self._listeners.remove(listener)
    except ValueError:  # thrown when the listener not in list
      pass


  def _notify_listeners(self, level=1):
    '''Notifies listeners that we've had a change'''
    # this gets called a lot, so the if statement is used to speed it up
    if len(self._listeners) > 0:
      for listener in self._listeners:
        listener(self, level)


  def _invalidate_indexes(self):
    '''Invalidates the indices already calculated on this table.  This occurs anytime
       data are modified in this table'''
    if len(self._indexes) > 0:
      self._indexes = {}
    if self.filterindex != None:
      self.filterindex = None
      
      
  def is_changed(self):
    '''Returns whether the table has been changed since loading'''
    return self.changed
    
    
  def set_changed(self, changed):
    '''Sets whether the class has been changed since loading.  This is not normally
       called by users.'''
    self.changed = changed

    
  def set_readonly(self, readonly_flag=False):
    '''Sets the read only status of this table.  Tables that are read only cannot
       be modified.  Normally, tables are initially not read only (i.e. can be modified).
       The only exception is tables loaded from databases, which are read only.
       
       @param readonly_flag: True or False, depending upon whether the table should be read only or not.
       @type  readonly_flag: bool
    '''
#    assert self._cursorlen < 0, 'Database relations are read-only by default.  Copy to a Picalo table to be able to modify it.'
    self.readonly = readonly_flag
    self._notify_listeners(level=1)
    

  def is_readonly(self):
    '''Returns whether this table is read only.
       @return:  Whether this table is read only.
       @rtype:   bool
    '''
    return self.readonly
    
    
  def index(self, *col_names):
    '''Returns an index on this table for the given column name(s).  This method
       allows you to find the record indices that match specific keys.  If only
       one column name is specified, the index will have as many records as there
       are unique records in the column.  If multiple columns are specified, the
       index will have as many records as there are unique combinations of the records
       in the columns.
       
       Indices are used throughout Picalo internally and are not normally accessed
       by users.  However, many analyses need to calculate indices directly, and exposing
       this method allows this behavior.
    
       This method can be called repeatedly without affecting performance. If 
       the table data have not been modified since the last time a specific index 
       was asked for, it uses the previously calculated index.
       
       @param col_names:   One or more column names (separated by commas) to calculate the unique index on.
       @type  col_names:   str
    '''
    check_valid_table(self, *col_names)
    cols = tuple(col_names)
    # make sure we have this index
    if not self._indexes.has_key(cols):
      d = {}
      for i in range(len(self)):
        if len(cols) == 1:
          key = self[i][cols[0]]
        else:
          key = tuple( [ self[i][col] for col in cols ] )
        if d.has_key(key):
          d[key].append(i)
        else:
          d[key] = [i]
      self._indexes[cols] = d
    # return it from the cache
    return self._indexes[cols]
    
  
  def find(self, **pairs):
    '''Finds the record indices with given key=record pairs.  This method
       is not normally used directly -- use Simple.select_by_record instead.
       
       This method is different than Simple.select_by_record in that it does not create 
       a new table.  Simple.select_by_record creates a new table consisting of copies of
       all matching records.  In contrast, this method simply returns the record indices 
       of the matching records.  
       
       This method *is* efficient and can be used often.  It calculates 
       indices as needed and should select very fast.
    
       Example:
        >>> table = Table([
        ...  ('col001', int),
        ...  ('col002', int),
        ...  ('col003', unicode),
        ... ],[
        ...  [5,6,'flo'], 
        ...  [3,2,'sally'], 
        ...  [4,6,'dan'], 
        ...  [4,7,'stu'], 
        ...  [4,7,'ben'],
        ...  [4,6,'benny'],
        ... ])
        >>> results = table.find(col001=6, col000=4)
        >>> results
        [2, 5]
  
       @param pairs:   column=record pairings giving the key(s) to select on.
       @type  pairs:   object
       @return:        A list of indices that match the pairs.
       @rtype:         list 
    '''       
    # no need to check type of pairs because self.index will do it
    idx = self.index(*pairs.keys())
    if len(pairs) == 1:
      vals = pairs.values()[0]
    else:
      vals = tuple(pairs.values())
    return idx.get(vals, [])
    
  
  def filter(self, expression=None):
    '''Filters the table with the given expression.  Until the filter is either
       replaced or cleared, only the records matching the filter will be available
       in the table.  All Picalo functions will see only this limited view of the
       table in their analyses.  In other words, it is as if the filtered record is
       not in the table at all until the filter is removed.
       
       Filters are transient.  They do not save with the table and they do not
       carry over if you copy a table.  They are simply temporary filters that you
       can use to restrict Picalo analyses to a few records.
       
       Only one filter can be active at any time.  Setting a new filter will replace
       the existing one.
       
       In creating your expression, use the standard record['col'] notation
       to access individual records in the table.
       
       @param expression: A valid Picalo expression that returns True or False.  If the
                          expression evaluates to Frue, the record is included in the filtered
                          view.  If False, the record is hidden from view.
       @type  expression: str
    '''
    self._invalidate_indexes()   # calls the update filter method
    if expression:
      self.filterexpression = PicaloExpression(expression)
    else:
      self.filterexpression = None
    self._update_filter_index()
      
      
  def clear_filter(self):
    '''Clears any active filter on this table, restoring the view to all
       records in the table'''
    self.filter(None)
    
    
  def is_filtered(self):
    '''Returns True if this table has an active filter'''
    return self.filterexpression != None
    
    
  def get_filter_expression(self):
    '''Returns the filter expression as a PicaloExpression object, or None if no filter is applied.'''
    return self.filterexpression
    
  
  def _update_filter_index(self):
    '''Updates the filter index.'''
    if self.filterexpression:
      newfilterindex = []
      for i in range(self.record_count(respect_filter=False)):
        rec = self.record(i, respect_filter=False)
        val = self.filterexpression.evaluate([{'record':rec, 'recordindex':i}, rec])
        if isinstance(val, error):
          self.clear_filter()
          raise AssertionError, 'Error filtering record ' + str(i) + ': ' + str(val.get_message())
        elif val == True:
          newfilterindex.append(i)
      self.filterindex = newfilterindex
    else:
      self.filterindex = None
    self._notify_listeners(level=2)
    
      
  def _get_filtered_index(self, index):
    '''Returns the filtered index of the given index, if a filter is in place.
       This method is not thread-safe.
    '''
    if self.filterexpression:
      # do we need to recreate the index?
      if self.filterindex == None:
        self._update_filter_index()
      # return the filtered index   
      return self.filterindex[index]
    return index
    
    
  def _calculate_columns_map(self):
    '''Calculates the columns map based upon the current columns.
       This method should not be called externally.
    '''
    self.columns_map_names = PersistentDict()
    self.columns_map_virtual = PersistentDict()
    self.num_columns = 0
    for i in range(len(self.columns)):
      if self.columns[i].virtual_col_index >= 0:
        self.columns_map_names[self.columns[i].name] = i
        self.columns_map_virtual[self.columns[i].virtual_col_index] = i
        self.num_columns += 1
    self._invalidate_indexes()
    
    
  def __copy__(self):
    '''Returns a copy of this table.  Use this method to make a full copy of all data.
       The normal way to copy data (and invoke this metho) is with the [:] syntax.
       
       Example:
        >>> table1 = Table([
        ...  ('col001', int),
        ...  ('col002', int),
        ...  ('col003', unicode),
        ... ],[
        ...  [5,6,'flo'], 
        ...  [3,2,'sally'], 
        ...  [4,6,'dan'], 
        ...  [4,7,'stu'], 
        ...  [4,7,'ben'],
        ...  [4,6,'benny'],
        ... ])
        >>>
        >>> table2 = table1[:]
        >>> # table2 now has the same columns and data as table1, but it is a 
        >>> # new table -- data changes to one don't affect the other.
        >>>
        >>> table3 = table1
        >>> # table3 is now a second name for table1 -- they are the same table.
        >>> # any changes to table3 will show in table1 and vice versa
    
       @return:  A new Table object with the columns and data of this Table.
       @rtype:   returns
    '''
    return Table(self.columns, self)    


  def _populate_record(self, rec, *a, **k):
    '''Populates a record with data.
       Examples:
         - Format 1:  mytable._populate_record(val1, val2)
         - Format 2:  mytable._populate_record([val1, val2])
         - Format 3:  mytable._populate_record(head1=val1, head2=val2)
         - Format 4:  mytable._populate_record({'head1':val1, 'head2':val2})
         - Format 5:  mytable._populate_record({0:val1, 1:val2})
       You cannot mix formats in the same call.
       
       This method is used internall and should not be called by user code.
    '''
    # set the parameters
    if len(k) > 0:  # Format 3
      for key, record in k.items():
        rec[key] = record

    elif len(a) > 0 and isinstance(a[0], (types.ListType, types.TupleType, Record)):  # format 2 above
      for row in a: # in case more than one list was given
        for i in range(self.column_count()):
          try:
            rec[i] = row[i]
          except IndexError:  # we didn't have enough input values for the number of columns, so just leave None
            pass
          
    elif len(a) > 0 and isinstance(a[0], types.DictType):  # formats 4 & 5 above
      for row in a: # in case more than one dict was given
        for key, record in row.items():
          rec[key] = record

    else:   # assume format 1 above
      for i in range(self.column_count()):
        try:
          rec[i] = a[i]
        except IndexError:  # we didn't have enough input values for the number of columns, so just leave None
          pass
    

  ##### COLUMN METHODS #####
    
  def deref_column(self, col):
    '''Dereferences the col name to its index (if it is a name).  For example,
       if the column name "id" is given, it returns the index of this
       column (such as 0, 1, 2, etc.).
       
       The column can be specified as the column index, the column name,
       or even a negative index (from the last column backward).
       
       @param col_name:    The name of the column
       @type  col_name:    str/int
       @return:            The index of the column
       @rtype:             int
    '''
    ###   START OF NEAR DUPLICATE CODE   ###
    # There is duplicate code in Record.__get_item__ and Record.__set_item__.
    # If you change something here, modify both of those locations as well.
    # This method is called so much that the duplicate code speeds things up by several orders.
    
    # if a name, convert to the index
    if not isinstance(col, int):
      if not col in self.columns_map_names:
        raise IndexError('This table has no column named %s' % col)
      col = self.columns_map_names[col]
  
    # the column was referenced as a number
    else:
      origcol = col
      # if a negative index, convert to a positive one
      if col < 0:
        col += self.num_columns

      # ensure that we have enough columns to fulfill the request
      if col >= self.num_columns or col < 0:
        raise IndexError('Column index [%s] is out of range.  The table only has %s columns, referenced with [0] to [%s] or [-1] to [-%s].' % (origcol, self.num_columns, self.num_columns - 1, self.num_columns))

      # deref to the actual column location in the table
      col = self.columns_map_virtual[col]
    
    # return the column number
    return col

    ###   END OF NEAR DUPLICATE CODE   ###  


  def column(self, col):
    '''Returns a single column of the table.  Column records can be read 
       and modified but not deleted.  This is a useful method when
       you want to work with a single column of a table.
       
       The returned Column object is essentially a list of records 
       and can be treated as such.  Any function that takes a list
       can take a Column object.
       
       This method is used often in analyses.
       
       @param col:   The column name to return
       @type  col:   str
       @return:      A Column object representing the specified column of the table.
       @rtype:       Column
    '''
    check_valid_table(self, col)
    return self.columns[self.deref_column(col)]
   
    
  def get_columns(self):
    '''Returns the column objects of this table.  Columns are Column objects, which
       contain the column name, calculated expression (if any), type if set, etc.
       
       Most users should call get_column_names() instead as it only returns
       the names of the columns.  Only call this method if you need to access
       the internal column objects.
    
       @return:     A list of Column objects
       @rtype:      list
    '''
    return [ self.columns[self.columns_map_virtual[i]] for i in range(self.column_count()) ]
    
    
  def get_column_names(self):
    '''Returns the column names of this table.  
       
       @return:     A list of string names of columns
       @rtype:      list
    '''
    return [ self.columns[self.columns_map_virtual[i]].name for i in range(self.column_count()) ]
    
    
  def column_count(self):
    '''Retrieves the number of columns in a table.  Note that since
       Picalo lists are zero-based, you access individual columns
       starting with 0.  
       
       So although column_count() may report 3 columns, you access
       them via table.column(0), table.column(1), and table.column(2).
       
       However, using column names rather than direct indices is easier
       and more readable.  table['colname'] returns the given column.
    
       @return:     The number columns in the table
       @rtype:      int
    '''
    return self.num_columns
    
  
  def set_name(self, col, name):
    '''Changes the name of an existing column.  The column name must be a
       valid Picalo name and must be unique to other column names in the table.'''
    self.column(col).set_name(name)
    
    
  def set_type(self, col, col_type=None, format=None, expression=None):
    '''Sets the type of a column.   The type must be a valid <type> object,
       such as int, float, str, unicode, DateTime, etc.  All records in this
       column will be converted to this new type.
       
       The format is an optional format to be used for printing the record
       and showing the record in the Pialo GUI.
       
       @param col:          The column name or index to set the type of
       @type  col:          str
       @param col_type:     The new type of this column.
       @type  col_type:     type
       @param format:       A Picalo expression that evaluates to a string
       @type  format:       str
       @param expression:   A Picalo expression that calculates this column.
    '''
    check_valid_table(self, col)
    return self.column(col).set_type(col_type, format, expression)


  def set_format(self, col, format=None):
    '''Sets the format of this column.  The format is used for printing the record 
       and showing the record in the Picalo GUI.
       
       The format should be a Picalo expression that evaluates to a string.  Use the
       'record' variable for the record of the current cell.
       
       Note that this is not an input mask.  It doesn't affect the internal record of 
       the field records.  It only affects how it is displayed on the screen.
       
       Example:
         # shows the current record in uppercase
         table.set_format('Salary', "record.upper()")
       
       @param col:     The column name or index to set the type of
       @type  col:     str
       @param format:  A Picalo expression that evaluates to a string
       @type  format:  str
    '''
    check_valid_table(self, col)
    return self.column(col).set_format(format)

  
  def _insert_column_(self, index, coldef, records=None, expression=None):
    '''Internal method to inesrt a column once the Column object is created'''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    assert isinstance(index, (types.IntType, types.StringType, types.UnicodeType)), 'Invalid column index: ' + str(index)
    if isinstance(index, types.IntType):
      assert index <= self.column_count() and index >= 0, 'Invalid column index: ' + str(index)
    assert isinstance(records, (types.NoneType, types.ListType, types.TupleType, Record, Table)), 'Please specify the new column records as a list.'
    # if the records variable is a table, use the first column of the table
    if isinstance(records, Table):
      records = records.column(0)

    # adjust the virtual_col_index numbers and then add to the columns list
    for c in self.columns:
      if c.virtual_col_index >= index:
        c.virtual_col_index += 1
    coldef.virtual_col_index = index
    self.columns.append(coldef)
    self._calculate_columns_map()
    
    # calculate the expression if we have one
    if expression:  # this is a static expression; an active one would have been placed in coldef
      pe = PicaloExpression(expression)
      coldef.static_expression = expression
      try:
        for i, rec in enumerate(self):
          show_progress('Calculating static expression in column...', float(i) / len(self))
          rec[coldef.name] = pe.evaluate([{'record': rec, 'recordindex': i}, rec])
      finally:
        clear_progress()      

    # set the data in the table if we were given data records
    if records:
      try:
        for i, rec in enumerate(self):
          show_progress('Adding data to new column...', float(i) / len(self))
          if i < len(records):
            val = records[i]
            if coldef.column_type:
              try:
                val = coldef.column_type(val)
              except Exception, e:
                val = error(e)
            rec[coldef.name] = val
      finally:
        clear_progress()      

    # update the table maps and indices
    self._notify_listeners(level=2)
    self.set_changed(True)
    return coldef
    
    
  def replace_column_values(self, column, expression):
    '''Replaces the given column with values calculated from the given expression.
       The previous values in this column are permanently deleted.
       
       This is similar to append_calculated_static(), except it modifies an existing
       column instead of adding a new column.
       
       Example 1:
         # replaces the values in the amount column with the number 10,000
         table.replace_column_values('amount', 10000)
         
       Example 2:
         # doubles the values in the amount column
         table.replace_column_values('amount', 'amount * 2')
      
       Example 3:
         # sets the values in the initials column with the first letter of each record's name
         table.replace_column_values('initials', 'fname[:1] + mname[:1] + lname[:1]')

       @param column:  The column name or index to replace the values in
       @type  column:  str
       @param format:  A Picalo expression used to set the values of the cell
       @type  format:  str
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    assert isinstance(column, (types.IntType, types.StringType, types.UnicodeType)), 'Invalid column index: ' + str(index)
    # dereference to the actual column index
    coldef = self.column(column)
    coldef.static_expression = expression
    # go through the records one by one and change them
    pe = PicaloExpression(expression)
    try:
      for i, rec in enumerate(self):
        show_progress('Replacing values...', float(i) / len(self))
        rec[column] = pe.evaluate([{'record': rec, 'recordindex': i}, rec])
    finally:
      clear_progress()      
    
    
  def append_column(self, name, column_type, records=None):
    '''Adds a new column to the table, optionally setting records of the new cells.

       @param name:        The new column name
       @type  name:        str
       @param column_type: The new column type (int, float, DateTime, unicode, str, etc)
       @type  column_type: type
       @param records:     A list of records to place into the cells of the new column (if a full Table, the first column is used)
       @type  records:     list, Column, or Table
       @return:            The new column object.
       @rtype:             Column
    '''
    return self._insert_column_(self.column_count(), Column(self, -1, name, column_type=column_type), records)

    
  def insert_column(self, index, name, column_type, records=None):
    '''Inserts a new column in the table at the given index location.

       @param name:        The new column name
       @type  name:        str
       @param column_type: The new column type (int, float, DateTime, unicode, str, etc)
       @type  column_type: type
       @param records:     A list of records to place into the cells of the new column (if a full Table, the first column is used)
       @type  records:     list, Column, or Table
       @return:            The new column object.
       @rtype:             Column
    '''
    return self._insert_column_(index, Column(self, -1, name, column_type=column_type), records)

    
  def append_calculated(self, name, column_type, expression):
    '''Adds a new, calculated column with records given by expression.
       Calculated columns act as regular columns in all ways.
       Their records are 'active' meaning their records change when the result
       of the expression.  In other words, they are recalculated each time they
       are used rather than being stored statically.  This is similar to the way
       Excel functions always reflect the most updated data records.
       
       Example: 
         >>> table = Table([('id', int)], [[1],[2],[4]])
         >>> table.append_calculated('plusone', "col000+1")
         >>> table.view()
         +--------+---------+
         | col000 | plusone |
         +--------+---------+
         |      1 |       2 |
         |      2 |       3 |
         |      4 |       5 |
         +--------+---------+
       
       @param name:        The new column name
       @type  name:        str
       @param column_type: The new column type (int, float, DateTime, unicode, str, etc)
       @type  column_type: type
       @param expression:  An expression that returns the record of the new field.  As shown in the example, use rec to denote the current record being evaluated.
       @type  expression:  str
       @return:            The new column object.
       @rtype:             Column
    '''
    return self.insert_calculated(self.column_count(), name, column_type, expression)
    
    
  def insert_calculated(self, index, name, column_type, expression):
    '''Inserts a new, calculated column with records given by expression at the given index location.
       Calculated columns act as regular columns in all ways.
       Their records are 'active' meaning their records change when the result
       of the expression.  In other words, they are recalculated each time they
       are used rather than being stored statically.  This is similar to the way
       Excel functions always reflect the most updated data records.
       
       Example: 
         >>> table = Table([('id', int)], [[1],[2],[4]])
         >>> table.insert_calculated(0, 'plusone', "col000+1")
         >>> table.view()
         +---------+--------+
         | plusone | col000 |
         +---------+--------+
         |       2 |      1 |
         |       3 |      2 |
         |       5 |      4 |
         +---------+--------+

       @param index:       The index location of the new column.  Previous column indices are incremented one to make room for the new column.
       @type  index:       int
       @param name:        The new column name
       @type  name:        str
       @param column_type: The new column type (int, float, DateTime, unicode, str, etc)
       @type  column_type: type
       @param expression:  An expression that returns the record of the new field.  As shown in the example, use rec to denote the current record being evaluated.
       @type  expression:  str
       @return:            The new column object.
       @rtype:             Column
    '''
    return self._insert_column_(index, Column(self, -1, name, column_type, expression=expression))
  
  
  def append_calculated_static(self, name, column_type, expression):
    '''Adds a new, calculated column with records given by expression.
       The records are calculated immediately using expression, and then they are static.
       In other words, this method calculates a new, regular column.  The records
       are not 'active' in the sense that append_calculated() columns are active.
       When this method returns, the new column is the same as any other, non-calculated
       column.
       
       Example: 
         >>> table = Table([('id', int)], [[1],[2],[4]])
         >>> table.append_calculated('plusone', int, "col000+1")
         >>> table.view()
         +--------+---------+
         | col000 | plusone |
         +--------+---------+
         |      1 |       2 |
         |      2 |       3 |
         |      4 |       5 |
         +--------+---------+
       
       @param name:        The new column name
       @type  name:        str
       @param column_type: The type of the new column (str, Date, int, long, etc.)
       @type  column_type: type
       @param expression:  An expression that returns the record of the new field.  As shown in the example, use rec to denote the current record being evaluated.
       @type  expression:  str
       @return:            The new column object.
       @rtype:             Column
    '''
    return self.insert_calculated_static(self.column_count(), name, column_type, expression)
    
    
  def insert_calculated_static(self, index, name, column_type, expression):
    '''Inserts a new, calculated column with records given by expression at the given index location.
       The records are calculated immediately using expression, and then they are static.
       In other words, this method calculates a new, regular column.  The records
       are not 'active' in the sense that append_calculated() columns are active.
       When this method returns, the new column is the same as any other, non-calculated
       column.

       Example: 
         >>> table = Table([('id', int)], [[1],[2],[4]])
         >>> table.append_calculated('plusone', int, "col000 + 1")
         >>> table.view()
         +--------+---------+
         | col000 | plusone |
         +--------+---------+
         |      1 |       2 |
         |      2 |       3 |
         |      4 |       5 |
         +--------+---------+
       
       @param index:       The index location of the new column.  Previous column indices are incremented one to make room for the new column.
       @type  index:       int
       @param name:        The new column name
       @type  name:        str
       @param column_type: The type of the new column (str, Date, int, long, etc.)
       @type  column_type: type
       @param expression:  An expression that returns the record of the new field.  As shown in the example, use rec to denote the current record being evaluated.
       @type  expression:  str
       @return:            The new column object.
       @rtype:             Column
    '''
    return self._insert_column_(index, Column(self, -1, name, column_type), expression=expression)


  def move_column(self, column, new_index):
    '''Moves a column to another location in the table.  A column
       can be moved in front of other columns or behind other columns
       with this method.  
       
       The column parameter is the name of the column to be moved,
       the index of the column to be moved, or the column object itself.
       
       The new_index parameter is the new index for this column.  This
       can be seen as the insertion point, or the column the moved column
       will be placed in front of.  It can be specified as a column index,
       a column name, or a column object.
       
       @param column: The name, index, or column object to be moved.
       @type  column: int/str/Column
       @param new_index: The name, index, or column object that the column will be placed before.
       @type  new_index: int/str/Column
    '''
    # this is not the fastest algorithm, but it's simple and still doesn't take much time
    coldef = self.column(column)
    if new_index != self.column_count():  # if it equals column count, we want to go to the end of the table
      new_index = self.column(new_index).virtual_col_index
    colnames = self.get_column_names()
    del colnames[coldef.virtual_col_index]
    if new_index > coldef.virtual_col_index:
      colnames.insert(new_index-1, coldef.name)
    else:
      colnames.insert(new_index, coldef.name)

    # now that we have the names in the right order, just call reorder_columns
    self.reorder_columns(colnames)
    
    
  def reorder_columns(self, columns):
    '''Reorders the columns according to the given list.  This is an alternative
       to move_column.  If you know the exact order you want the columns in, use
       this method to explicitly set them.
       
       The columns parameter is a list giving the new order.  Its items can
       be current column indices, names, or column objects.
       
       @param columns: A list giving the new column order (use column names, not indices).
       @type  columns: list
    '''
    # check the parameters
    assert isinstance(columns, (types.ListType, types.TupleType)), 'The columns parameter must be a list describing the new column order.'
    assert len(dict([ (self.deref_column(idx), idx) for idx in columns ])) == len(columns), 'You cannot specify a column twice in the columns parameter.'
    for col in columns:
      self.column(col)  # checks for valid name and range if int

    # first deactivate all the columns
    for coldef in self.columns:  
      coldef.virtual_col_index = -1
    
    # activate them in the order the user wants
    for i, col in enumerate(columns):
      self.column(col).virtual_col_index = i

    # notify people that we've changed
    self._calculate_columns_map()
    self.set_changed(True)
    self._notify_listeners(level=2)


  def delete_column(self, *columns):
    '''Removes a column from the table and discards the records.  
       Remaining column indices are decremented to reflect the new 
       table structure.  Column names (columns) are not modified.
       
       Example: 
         >>> table.delete_column('id')          # deletes the column named 'id' from table
         >>> table.delete_column('id', 'age')   # deletes the columns named 'id' and 'age' from table
       
       @param columns:    The name or index of the column to be removed.  The parameter can be specified more than once.
       @type  columns:    str or list
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    check_valid_table(self, columns)
    
    # go through and delete each name in the list
    indices = [ self.column(col).virtual_col_index for col in columns ]
    colnames = self.get_column_names()
    for name in [ self.column(c).name for c in columns ]:
      del colnames[colnames.index(name)]

    # reorder with the new names
    self.reorder_columns(colnames)
      
    
  ##### CONTAINER METHODS ####
  
  
  def record(self, index, respect_filter=True):
    '''Retrieves the record at the given index.  This is one of the most-used
       methods in the Picalo toolkit as it gives you access to records.
       
       In keeping with most computer languages, Picalo indices are always
       zero-based.  This may require a slight adjustment for some users,
       but it makes mathematical calculations much easier and has other
       implications. This means that record 1 is table[0], record 2 is 
       table[1], and so forth.
       
       Note that the shortcut way to access this method is the simple [n]
       notation, as in table[1] to access the second record.  table.record()
       is rarely called as the shortcut is preferred instead.
    
       For advanced users: Index can also be a slice, as in table[2:5] to 
       return a new Picalo table including only records 2, 3, and 4.  See 
       the Python documentation for more information on slices.
       
       The method respects any filters on the table by default.  This can be
       overridden to ignore the filter.
        
       Example:
         >>> table = Table([('id', int)], [[1],[2],[3],[4]])
         >>> table2 = table[1]
         >>> # table2 is now a Record object pointing at [2]
       
       Example 2:
         >>> table = Table([('id', int)], [[1],[2],[3],[4]])
         >>> print table[1]['col000']
         2
       
       @param index:   The zero-based index of the record to pull.
       @type  index:   int
       @return:        A Picalo Record object, which allows access to members via column name.
       @rtype:         Record
    '''
    assert isinstance(index, (types.IntType, types.LongType, types.SliceType)), 'Please specify the index of the record as an integer.'
    # if a slice, return a new picalo table
    # list would otherwise return a new list of Record objects
    if isinstance(index, types.SliceType):
      newtable = Table(self.get_columns())
      start, step, stop = index.start, index.step, index.stop
      if start == None: start = 0
      if step == None: step = 1
      if stop == None: stop = self.record_count(respect_filter)
      stop = min(stop, self.record_count(respect_filter))
      for i in range(start, stop, step):
        newtable.append(self.record(i, respect_filter))
      return newtable
    # if a negative number, convert it to positive
    if index < 0:
      index = self.record_count(respect_filter) + index
    # otherwise, get the record
    if respect_filter:
      index = self._get_filtered_index(index)
    return self._data[index]
    
  
  def __getitem__(self, index):
    '''Retrieves a record or a full column of the table.  If the given index is a column name,
       the column object is returned.  If the given index is an integer, the given 
       row object is returned.
    
       @param index:   The zero-based index of the record to pull OR the string name of the column to pull.
       @type  index:   int or str
       @return:        A Picalo Record object or Picalo Column object from this table
       @rtype:         Record or Column
    '''
    if isinstance(index, types.StringTypes):  # a column name
      return self.column(index)
    else:  # a record index
      return self.record(index)
    
  
  def record_count(self, respect_filter=True):
    '''Returns the number of records in this table.  The method
       respects any active filters on the table by default.

       @return:  The number of records in the table.
       @rtype:   int
    '''
    # do we need to recreate the index?
    if respect_filter and self.filterexpression:
      if self.filterindex == None:
        self._update_filter_index()
    if self.filterindex != None and respect_filter:
      return len(self.filterindex)
    return len(self._data)

    
  def __len__(self):
    '''Return the number of records in the table.  The method respects
       any active filters on the table.
    
       @return:  The number of records in the table.
       @rtype:   int
    '''
    return self.record_count()
    
    
  def __setitem__(self, index, record):
    '''Replaces an existing Record with the given data.  Use = to set items.
    
       Example:
         >>> table = Table([('id', int)], [[1,2], [3,4]])
         >>> table[1] = [5,6]
         >>> table.view()
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         |      1 |      2 |
         |      5 |      6 |
         +--------+--------+
       
       @param index: The index of the record where data will be replaced.
       @type  index: int
       @param record: A Record object, a list, or a Python sequence containing data for the 
       @type  record: Record/list/tuple of columns for the record records.
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    assert isinstance(index, (types.IntType, types.LongType, types.SliceType)), 'Please specify the index of the record as an integer.'
    for i in range(min(len(record), self.column_count())):
      self[index][i] = record[i]
    self._invalidate_indexes()
    self.set_changed(True)
    self._notify_listeners(level=1)
    
    
  def __iter__(self):
    '''Returns an interator to the Records in this Table.  This is normally achieved through:
         >>> for record in mytable:
         >>>   # do something with each record object
           
       @return:    An iterator to this Table.
       @rtype:     iterator
    '''
    return TableIterator(self)
    
    
  def iterator(self, respect_filter=True):
    '''Returns an iterator to the Records in thistTable.  This is normally achieved through:
         >>> for record in mytable:
         >>>   # do something with each record object
         
       This method is provided to allow you to ignore the filter if you want.  The 
       regular iterator syntax (for record in mytable) always respects any active filters.
           
       @return:    An iterator to this Table.
       @rtype:     iterator
    '''
    return TableIterator(self, respect_filter)
    
    
  def __delitem__(self, index):
    '''Removes a record from the table.  The proper use to call this method
       is 'del table[i]' where i is the record number to remove.  You can also
       specify columns to delete with 'del table["colname"]' where colname
       is the column name you want to delete.'''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    if isinstance(index, types.StringTypes):  # for column deletion
      return self.delete_column(index)
    assert isinstance(index, (types.IntType, types.LongType, types.SliceType)), 'Please specify the index of the record as an integer.'
    # call recursively if a slice type
    if isinstance(index, types.SliceType):
      start, stop, step = index.start, index.stop, index.step
      if stop >= 0:  # if stop is left off, it goes to maxint
        stop = min(stop, self.record_count())
      if not step:
        step = 1
      lowest = min(start, stop)
      for i in range(start, stop, step):
        del self[lowest]
      return
    # a regular deletion
    index = self._get_filtered_index(index)
    del self._data[index]
    self._invalidate_indexes()
    self.set_changed(True)
    self._notify_listeners(level=2)
    
    
    
  ######  LIST METHODS  ######
  
  def append(self, *a, **k):
    '''Inserts a new record at the end of this table.  This is the primary
       way to add new data to a table.  If you need to insert a row in the middle
       of a table, use the insert() method.
    
       Records can be added in any of the following ways:

       Format 1:
         - newrec = mytable3.append()
         - newrec['ID'] = 4
         - newrec['Name'] = 'Homer'
         - newrec['Salary'] = 15000

       Format 2:
         - mytable.append(4, 'Homer', 1500)

       Format 3:
         - mytable3.append([4, 'Homer', 1500])

       Format 4:
         - mytable3.append({'ID':5, 'Name':'Marge', 'Salary': 275000})

       Format 5:
         - mytable3.append({0:5, 1:'Marge', 2: 275000})

       Format 6:
         - mytable3.append(ID=5, Name='Krusty', Salary=50000)

       You cannot mix formats in the same call.
       
       @return:   The new record object (that was appended to the end of the table)
       @rtype:    Record
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    # append the record and adjust our ending index
    rec = Record(self)
    self._data.append(rec)
    self._populate_record(rec, *a, **k)
    # update indices
    self.set_changed(True)
    self._invalidate_indexes()
    self._notify_listeners(level=2)
    return rec
    
    
  def insert(self, *a, **k):
    '''Inserts a new record at the given index location.  The first parameter *must*
       be the index location to insert the record.  The remaining parameters are the
       same as the append() method.  Possible formats are (assuming you want to place
       the new record in the second row and push all existing records down one):

       Format 1:
         - newrec = mytable3.insert(2) # insert before row 2
         - newrec['ID'] = 4
         - newrec['Name'] = 'Homer'
         - newrec['Salary'] = 15000

       Format 2:
         - mytable.insert(2, 4, 'Homer', 1500) # insert before row 2

       Format 3:
         - mytable3.insert(2, [4, 'Homer', 1500]) # insert before row 2

       Format 4:
         - mytable3.insert(2, {'ID':5, 'Name':'Marge', 'Salary': 275000}) # insert before row 2

       Format 5:
         - mytable3.insert(2, {0:5, 1:'Marge', 2: 275000}) # insert before row 2

       Format 6:
         - mytable3.insert(2, ID=5, Name='Krusty', Salary=50000) # insert before row 2

       You cannot mix formats in the same call.
       
       @return:  The new record object (that was inserted to the end of the table)
       @rtype:   returns
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    rec = Record(self)
    index = a[0]
    if index < 0:
      index = len(self) + index
    if index < 0:
      raise IndexError, 'list index out of range'
    index = self._get_filtered_index(index)
    self._data.insert(index, rec)
    self._populate_record(rec, *a[1:], **k)
    # update indices    
    self.set_changed(True)
    self._invalidate_indexes()
    self._notify_listeners(level=2)
    return rec
    
    
  def extend(self, table):
    '''Appends the records in the given table to the end of this table.
       The two tables must have the same number of columns to be merged. 

       Example:
         >>> mytable3.extend(mytable2)  # appends mytable2 to the end of mytable3
         
       @param table:  The records of the specified table will be added to this table.
       @type  table:  Table
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    if not isinstance(table, Table):
      assert isinstance(table, (Table, types.TupleType, types.ListType)), 'Invalid table type given to extend.'
      for item in table:
        assert isinstance(item, (Record, types.TupleType, types.ListType, types.DictType)), 'Invalid table type given to extend.'
    try:
      totalrecs = float(len(table))
      for i, rec in enumerate(table):
        show_progress('Initializing records...', float(i) / totalrecs)
        self.append(rec)
    finally:
      clear_progress()
    self.set_changed(True)
    self._invalidate_indexes()
    self._notify_listeners(level=2)
    
      
  ######  NUMERIC METHODS  ########
  
  def __iadd__(self, other):
    '''This is essentially the extend method.  I included a += method directly
       to increase efficiency (so no intermediate table needs to be created)
    '''
    check_valid_table(other)
    self.extend(other)
    return self
  
  
  def __add__(self, other):
    '''Adds two tables together.
       To be added together, two tables should have the same number of columns
       and be compatible with one another.  
       The new table includes the records of the first table followed by the
       records of the second table.  The column columns (and column calculations)
       of the first table are carried to the new table.
       The two source tables are not modified.
       
       Stated differently, this method copies the first table and appends the second 
       table's records to it.
       
       Some users might expect this method to add individual cells together, similar to
       matrix addition.  This is *not* the case.  Instead, the new table includes all
       records from both tables.
       
       Example:
         >>> t1 = Table(2, [[1,1], [2,2]])
         >>> t2 = Table(2, [[3,3], [4,4]])
         >>> t3 = t1 + t2
         >>> t3.view()
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         |      1 |      1 |
         |      2 |      2 |
         |      3 |      3 |
         |      4 |      4 |
         +--------+--------+
       
       @param other:  A Table to be added to this one.
       @type  other:  Table
       @return:       A new table with records from both tables
       @rtype:        Table
    '''
    if not isinstance(other, Table):
      assert isinstance(other, (types.TupleType, types.ListType)), 'Invalid table type given to extend.'
      for item in other:
        assert isinstance(item, (types.TupleType, types.ListType)), 'Invalid table type given to extend.'
    table = Table(self.get_columns(), self) # adds the records of the first table
    table.extend(other)
    return table
    
    
  def __sub__(self, other):
    raise NotImplementedError, 'Table subtraction is not allowed.'


  def __eq__(self, other):
    '''Returns whether this table is equal to another table (or list of lists).  Only the records
       of the table are compared -- not the column names or column definitions.'''
    if isinstance(other, (Table, types.ListType, types.TupleType)):
      if len(self) == len(other):
        for i in range(len(self)):
          if self[i] != other[i]:
            return False
        return True
    return False
    
  def __ne__(self, other):
    '''Returns whether this table is not equal to another table (or list of lists). Only the records
       of the table are compared -- not the column names or column definitions.'''
    return not self.__eq__(other)
    
    
  def sort(self, cmp=None, key=None, reverse=False):
    '''Sorts this table with optional arguments.  This is the standard Python sort method that
       iterable objects have.  See the Python sorting HowTo for more information on sorting the
       Pythonic way.
       
       Most users will probably prefer the Simple.sort(table, ...) function for sorting.  See
       the Simple.sort documentation for more information on this method of sorting.
       
       Example 1:
         >>> t1 = Table(2, [[2,2], [1,1], [5,5], [0,0]])
         >>> t1.view()
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         | 2      | 2      |
         | 1      | 1      |
         | 5      | 5      |
         | 0      | 0      |
         +--------+--------+
         >>> t1.sort(key=lambda rec: rec.col000)  # to sort by the 'col000' column
         >>> t1.view()
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         | 0      | 0      |
         | 1      | 1      |
         | 2      | 2      |
         | 5      | 5      |
         +--------+--------+
         
       Example 2:
         >>> # this example uses Simple.sort to achieve the same result as Example 1
         >>> t1 = Table(2, [[2,2], [1,1], [5,5], [0,0]])
         >>> t1.view()
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         | 2      | 2      |
         | 1      | 1      |
         | 5      | 5      |
         | 0      | 0      |
         +--------+--------+
         >>> Simple.sort(t1, True, 'col001') # table name, ascending, sort col 1, sort col 2, ...
         >>> t1.view()         
         +--------+--------+
         | col000 | col001 |
         +--------+--------+
         | 0      | 0      |
         | 1      | 1      |
         | 2      | 2      |
         | 5      | 5      |
         +--------+--------+
   
       @param cmp:     An optional function that compares two items
       @type  cmp:     function
       @param key:     A function that takes a single item and returns a version of it for use in sorting.
       @type  key:     function
       @param reverse: Whether to sort in reverse
       @type  reverse: bool
    '''
    assert not self.is_readonly(), 'This table is set as read-only and cannot be modified.'
    self._data.sort(cmp, key, reverse)
    self.set_changed(True)


  def structure(self):
    '''Returns the structure of this table, including column names, input and output types,
       and general statistics.
    
       @return:   A Picalo table describing the structure of this table.
       @rtype:    Table
    '''
    struct = Table([('Column', unicode), ('Type', str), ('Expression', str), ('Format', unicode)])
    for col in self.columns:
      rec = struct.append()
      rec['Column'] = col.name
      rec['Type'] = str(col.column_type)
      rec['Expression'] = col.expression and col.expression.expression or None
      rec['Format'] = col.format
    return struct

   
  ######  PRINTING METHODS  #######
  
  def prettyprint(self, fp=None, center_columns=True, space_before=' ', space_after=' ', col_separator_char='|', row_separator_char='-', join_char='+', line_ending=os.linesep, none='<N>', encoding='utf-8', respect_filter=True):
    '''Pretty prints the table to the given fp.  Note that the preferred way to print
       a table is to call "table.view()", which opens the table in the Picalo
       GUI if possible, or uses view() if in console mode.  In other words,
       you should normally use view() rather than this method.
       
       @param fp:                  An open file pointer object.  If None, defaults to standard output stream.
       @type  fp:                  file
       @param center_columns:      Whether to center columns or not.
       @type  center_columns:      bool
       @param space_before:        An optional spacing between the leading column separator and the field record.
       @type  space_before:        str
       @param space_after:         An optional spacing between the field and the trailing column separator.
       @type  space_after:         str
       @param col_separator_char:  An optional column separator character to use in the printout.
       @type  col_separator_char:  str
       @param row_separator_char:  An optional row separator character to use in the printout.
       @type  row_separator_char:  str
       @param join_char:           An optional character to use when joining rows and columns.
       @type  join_char:           str
       @param line_ending:         An optional line ending character(s) to use.
       @type  line_ending:         str
       @param none:                The record to print when cells are set to the special None record.
       @type  none:                str
       @param encoding:     The unicode encoding to write with.  This should be a value from the codecs module.  If None, the encoding is guessed to utf_8, utf-16, utf-16-be, or utf-16-le
       @type  encoding:     str
    '''
    assert isinstance(space_before, types.StringTypes), 'The space_before parameter must be a string.'
    assert isinstance(space_after, types.StringTypes), 'The space_after parameter must be a string.'
    assert isinstance(col_separator_char, types.StringTypes), 'The col_separator_char parameter must be a string.'
    assert isinstance(row_separator_char, types.StringTypes), 'The row_separator_char parameter must be a string.'
    assert isinstance(join_char, types.StringTypes), 'The join_char parameter must be a string.'
    assert isinstance(line_ending, types.StringTypes), 'The line_ending parameter must be a string.'
    assert isinstance(none, types.StringTypes), 'The none parameter must be a string.'
    
    # Note to programmer: this method is used by save_fixed below, so be sure to maintain compatability
    # I set fp=None above because if fp=sys.stdout, it sets that at class creation time
    # and can't be redirected to the shell
    if fp == None:
      fp = sys.stdout 
      
    # encode for unicode output
    fp = codecs.EncodedFile(fp, 'utf_8', encoding)
    
    # calculate the maximum width of each column
    widths = [ 0 for i in range(self.column_count()) ]
    for record in self.iterator(respect_filter):
      for i in range(self.column_count()):
        field = record[i]
        if field == None:
          field = make_unicode(none)
        else:
          field = make_unicode(self.column(i).format_value(field))
        widths[i] = max(widths[i], len(field))
    for i in range(self.column_count()):
      widths[i] = max(widths[i], len(str(self.column(i).name)))
        
    # create the separator line
    linesep = make_unicode(join_char)
    for w in widths:
      linesep += make_unicode(row_separator_char)
      for ch in range(w):
        linesep += make_unicode(row_separator_char)
      linesep += make_unicode(row_separator_char) + make_unicode(join_char)
    linesep += make_unicode(line_ending)
        
    # print the headings
    if linesep != line_ending:
      fp.write(linesep)
    line = make_unicode(col_separator_char)
    for i in range(self.column_count()):
      if center_columns:
        line += make_unicode(space_before) + make_unicode(self.column(i).name).center(widths[i]) + make_unicode(space_after) + make_unicode(col_separator_char)
      else:
        line += make_unicode(space_before) + make_unicode(self.column(i).name).ljust(widths[i]) + make_unicode(space_after) + make_unicode(col_separator_char)
    line += make_unicode(line_ending)
    fp.write(line)
    if linesep != line_ending:
      fp.write(linesep)
        
    # print the rows
    for record in self.iterator(respect_filter):
      line = make_unicode(col_separator_char)
      for i in range(self.column_count()):
        field = record[i]
        if field == None:
          line += make_unicode(space_before) + make_unicode(none).center(widths[i]) + make_unicode(space_after) + make_unicode(col_separator_char)
        elif isinstance(field, (types.IntType, types.LongType, types.FloatType)):  # rjustification
          line += make_unicode(space_before) + make_unicode(self.column(i).format_value(field)).rjust(widths[i]) + make_unicode(space_after) + make_unicode(col_separator_char)
        else:
          line += make_unicode(space_before) + make_unicode(self.column(i).format_value(field)).ljust(widths[i]) + make_unicode(space_after) + make_unicode(col_separator_char)
      line += make_unicode(line_ending)
      fp.write(line.encode(encoding))
    
    # print the bottom table border
    if linesep != line_ending:
      fp.write(linesep)
    
    fp.flush()
    
    
  def view(self):
    '''Opens a spreadsheet-view of the table if Picalo is being run in GUI mode.
       If Picalo is being run in console mode, it redirects to prettyprint().
       This is the preferred way of viewing the data in a table.
    '''
    view(self)
    
    
  def __repr__(self):
    '''For debugging.  Use view() for a formatted printout.
       
       @return:   The number of rows and columns in the table.  
       @rtype:    str
    '''
    return '<Table: ' + str(len(self)) + ' rows x ' + str(self.column_count()) + ' cols>'
    
        
  ######  EXPORTING METHODS #######


  def save_delimited(self, filename, delimiter=',', qualifier='"', line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False):
    '''Saves this table to a delimited text file.  This method
       allows the specification of different types of delimiters
       and qualifiers.
       
       This method is for advanced users.  Most users should call
       save_tsv, save_csv, or save_fixed to save using one of the 
       accepted text formats.  See these methods for more information.
    
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param delimiter:    A field delimiter character, defaults to a comma (,)
       @type  delimiter:    str
       @param qualifier:    A qualifier to use when delimiters exist in field records, defaults to a double quote (")
       @type  qualifier:    str
       @param line_ending:  A line ending to separate rows with, defaults to os.linesep (\n on Unix, \r\n on Windows)
       @type  line_ending:  str
       @param none:         An parameter specifying what to write for cells that have the None value, defaults to an empty string ('')
       @type  none:         str
       @param encoding:     The unicode encoding to write with.  This should be a value from the codecs module, defaults to 'utf-8'.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    Filer.save_delimited(self, filename, delimiter=',', qualifier='"', line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False)
    
    
  def save_csv(self, filename, line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False):
    '''Saves this table to a Comma Separated Values (CSV) text file.
       CSV is an industry-standard way of transferring data between
       applications.  This is the preferred way of exporting data from
       Picalo.
       
       Note that although this is the preferred export method, it has some
       limitations.  These are limitations of the format rather than
       limitations of Picalo:
        - No type information is saved to the file.  All data is essentially turned into strings.
        - Be sure to use the correct encoding if using international languages.
        - Different standards for CSV exist (that's the nice thing about standards :).  This export uses the Microsoft version.
       
       Note that Microsoft Office seems to like CSV files better than TSV files.
    
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None record.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    Filer.save_csv(self, filename, line_ending=line_ending, none=none, encoding=encoding, respect_filter=respect_filter)
    
    
  def save_tsv(self, filename, line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False):
    '''Saves this table to a Tab Separated Values (TSV) text file.
       TSV is an industry-standard way of transferring data between
       applications.
       
       Note that although this is the preferred export method, it has some
       limitations.  These are limitations of the format rather than
       limitations of Picalo:
        - No type information is saved to the file.  All data is essentially turned into strings.
        - Be sure to use the correct encoding if using international languages.
        - Different standards for TSV exist (that's the nice thing about standards :).  This export uses the Microsoft version.
       
       Note that Microsoft Office seems to like CSV files better than TSV files.
    
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None record.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
    '''
    Filer.save_tsv(self, filename, line_ending=line_ending, none=none, encoding=encoding, respect_filter=respect_filter)
   
   
  def save_fixed(self, filename, line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False):
    '''Saves this table to a fixed width text file.  Fixed width
       files pad column records with extra spaces so they are easier
       to read.
       
       You should normally prefer CSV and TSV export formats to fixed
       as they have less limitations.  Fixed format was widely used
       in the early days of computers, and some servers still use the
       format.
    
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None record.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    Filer.save_fixed(self, filename, line_ending=os.linesep, none='', encoding='utf-8', respect_filter=False)
    
    
  def save_xml(self, filename, line_ending=os.linesep, indent='\t', compact=False, none='', encoding='utf-8', respect_filter=False):
    '''Saves this table to an XML file using a pre-defined schema.  If you need
       to save to a different schema, use the xml.dom.minidom class directly.
       
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with (when compact is False)
       @type  line_ending:  str
       @param indent:       The character(s) to use for indenting (when compact is False)
       @type  indent:       str
       @param compact:      Whether to compact the XML or make it "pretty" with whitespace
       @type  compact:      bool
       @param none:         An optional parameter specifying what to write for cells that have the None record.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''    
    Filer.save_xml(self, filename, line_ending=os.linesep, indent='\t', compact=False, none='', encoding='utf-8', respect_filter=False)
    
   
  def save_excel(self, filename, none='', respect_filter=False):
    '''Saves this table to a Microsoft Excel 97+ file.
       
       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param none:         An optional parameter specifying what to write for cells that have the None record.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''    
    Filer.save_excel(self, filename, none='', respect_filter=False)


  def save(self, filename, respect_filter=False):
    '''Saves this table in native Picalo format.  This is the preferred
       format to save tables in because all column types, formulas, and so
       forth are saved.
       
       @deprecated: Add tables to Project objects and load/save the project.  Loading/saving tables directly (the .pco format) is no longer supported.
    
       @param filename: The filename to save to.  This can also be an open stream.
       @type  filename: str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    Filer.save(self, filename, respect_filter=respect_filter)
    



# I can't import Filer until now because Table is not declared until now
import Filer


###############################################
###   View code for tables

def view(table):
  '''Opens a spreadsheet-view of the table if Picalo is being run in GUI mode.
     If Picalo is being run in console mode, it redirects to prettyprint().
     This is the preferred way of viewing the data in a table.
  '''
  # NOTE: this method is also used by Database.Query
  from picalo import mainframe
  if mainframe != None:
    # try to figure out the variable name of this table by inspecting the call stack
    f = inspect.currentframe().f_back.f_back
    locals = f.f_locals
    for name in f.f_code.co_names:
      if locals.has_key(name) and locals[name] == table:
        mainframe.openTable(name)
        return
  # if we get here, either we're in console mode or we couldn't find the variable name
  table.prettyprint()


  
##################################################################
###   An iterator for tables that respects the filters on them.
      
def TableIterator(table, respect_filter=True):     
  '''Returns a generator object to iterate over a table'''
  index = 0
  numrows = table.record_count(respect_filter)
  while index < numrows:
    yield table.record(index, respect_filter)
    index += 1       
      


  
  
  
  
  
  
  
  
  
  
  
  



### $ANTLR 2.7.5 (20050128): "excel-formula.g" -> "ExcelFormulaParser.py"$
### import antlr and other modules ..
import sys
import antlr

version = sys.version.split()[0]
if version < '2.2.1':
    False = 0
if version < '2.3':
    True = not False
### header action >>> 
__rev_id__ = """$Id: ExcelFormulaParser.py,v 1.4 2005/08/14 06:40:23 rvk Exp $"""

import struct
import Utils
from UnicodeUtils import upack1
from ExcelMagic import *
### header action <<< 
### preamble action>>>

### preamble action <<<

### import antlr.Token 
from antlr import Token
### >>>The Known Token Types <<<
SKIP                = antlr.SKIP
INVALID_TYPE        = antlr.INVALID_TYPE
EOF_TYPE            = antlr.EOF_TYPE
EOF                 = antlr.EOF
NULL_TREE_LOOKAHEAD = antlr.NULL_TREE_LOOKAHEAD
MIN_USER_TYPE       = antlr.MIN_USER_TYPE
TRUE_CONST = 4
FALSE_CONST = 5
STR_CONST = 6
NUM_CONST = 7
INT_CONST = 8
NAME = 9
EQ = 10
NE = 11
GT = 12
LT = 13
GE = 14
LE = 15
ADD = 16
SUB = 17
MUL = 18
DIV = 19
POWER = 20
PERCENT = 21
LP = 22
RP = 23
LB = 24
RB = 25
COLON = 26
COMMA = 27
SEMICOLON = 28
CONCAT = 29
REF2D = 30

class Parser(antlr.LLkParser):
    ### user action >>>
    ### user action <<<
    
    def __init__(self, *args, **kwargs):
        antlr.LLkParser.__init__(self, *args, **kwargs)
        self.tokenNames = _tokenNames
        ### __init__ header action >>> 
        self.rpn = ""
        ### __init__ header action <<< 
        
    def formula(self):    
        
        pass
        self.expr("V")
    
    def expr(self,
        arg_type
    ):    
        
        pass
        self.prec0_expr(arg_type)
        while True:
            if ((self.LA(1) >= EQ and self.LA(1) <= LE)):
                pass
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in [EQ]:
                    pass
                    self.match(EQ)
                    op = struct.pack('B', ptgEQ)
                elif la1 and la1 in [NE]:
                    pass
                    self.match(NE)
                    op = struct.pack('B', ptgNE)
                elif la1 and la1 in [GT]:
                    pass
                    self.match(GT)
                    op = struct.pack('B', ptgGE)
                elif la1 and la1 in [LT]:
                    pass
                    self.match(LT)
                    op = struct.pack('B', ptgLT)
                elif la1 and la1 in [GE]:
                    pass
                    self.match(GE)
                    op = struct.pack('B', ptgGE)
                elif la1 and la1 in [LE]:
                    pass
                    self.match(LE)
                    op = struct.pack('B', ptgLE)
                else:
                        raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                    
                self.prec0_expr(arg_type)
                self.rpn += op
            else:
                break
            
    
    def prec0_expr(self,
        arg_type
    ):    
        
        pass
        self.prec1_expr(arg_type)
        while True:
            if (self.LA(1)==CONCAT):
                pass
                pass
                self.match(CONCAT)
                op = struct.pack('B', ptgConcat)
                self.prec1_expr(arg_type)
                self.rpn += op
            else:
                break
            
    
    def prec1_expr(self,
        arg_type
    ):    
        
        pass
        self.prec2_expr(arg_type)
        while True:
            if (self.LA(1)==ADD or self.LA(1)==SUB):
                pass
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in [ADD]:
                    pass
                    self.match(ADD)
                    op = struct.pack('B', ptgAdd)
                elif la1 and la1 in [SUB]:
                    pass
                    self.match(SUB)
                    op = struct.pack('B', ptgSub)
                else:
                        raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                    
                self.prec2_expr(arg_type)
                self.rpn += op
            else:
                break
            
    
    def prec2_expr(self,
        arg_type
    ):    
        
        pass
        self.prec3_expr(arg_type)
        while True:
            if (self.LA(1)==MUL or self.LA(1)==DIV):
                pass
                la1 = self.LA(1)
                if False:
                    pass
                elif la1 and la1 in [MUL]:
                    pass
                    self.match(MUL)
                    op = struct.pack('B', ptgMul)
                elif la1 and la1 in [DIV]:
                    pass
                    self.match(DIV)
                    op = struct.pack('B', ptgDiv)
                else:
                        raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                    
                self.prec3_expr(arg_type)
                self.rpn += op
            else:
                break
            
    
    def prec3_expr(self,
        arg_type
    ):    
        
        pass
        self.prec4_expr(arg_type)
        while True:
            if (self.LA(1)==POWER):
                pass
                pass
                self.match(POWER)
                op = struct.pack('B', ptgPower)
                self.prec4_expr(arg_type)
                self.rpn += op
            else:
                break
            
    
    def prec4_expr(self,
        arg_type
    ):    
        
        pass
        self.prec5_expr(arg_type)
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in [PERCENT]:
            pass
            self.match(PERCENT)
            self.rpn += struct.pack('B', ptgPercent)
        elif la1 and la1 in [EOF,EQ,NE,GT,LT,GE,LE,ADD,SUB,MUL,DIV,POWER,RP,SEMICOLON,CONCAT]:
            pass
        else:
                raise antlr.NoViableAltException(self.LT(1), self.getFilename())
            
    
    def prec5_expr(self,
        arg_type
    ):    
        
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in [TRUE_CONST,FALSE_CONST,STR_CONST,NUM_CONST,INT_CONST,NAME,LP,REF2D]:
            pass
            self.primary(arg_type)
        elif la1 and la1 in [SUB]:
            pass
            self.match(SUB)
            self.primary(arg_type)
            self.rpn += struct.pack('B', ptgUminus)
        else:
                raise antlr.NoViableAltException(self.LT(1), self.getFilename())
            
    
    def primary(self,
        arg_type
    ):    
        
        str_tok = None
        int_tok = None
        num_tok = None
        ref2d_tok = None
        ref2d1_tok = None
        ref2d2_tok = None
        name_tok = None
        func_tok = None
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in [TRUE_CONST]:
            pass
            self.match(TRUE_CONST)
            self.rpn += struct.pack("2B", ptgBool, 1)
        elif la1 and la1 in [FALSE_CONST]:
            pass
            self.match(FALSE_CONST)
            self.rpn += struct.pack("2B", ptgBool, 0)
        elif la1 and la1 in [STR_CONST]:
            pass
            str_tok = self.LT(1)
            self.match(STR_CONST)
            self.rpn += struct.pack("B", ptgStr) + upack1(str_tok.text[1:-1])
        elif la1 and la1 in [INT_CONST]:
            pass
            int_tok = self.LT(1)
            self.match(INT_CONST)
            self.rpn += struct.pack("<BH", ptgInt, int(int_tok.text))
        elif la1 and la1 in [NUM_CONST]:
            pass
            num_tok = self.LT(1)
            self.match(NUM_CONST)
            self.rpn += struct.pack("<Bd", ptgNum, float(num_tok.text))
        elif la1 and la1 in [LP]:
            pass
            self.match(LP)
            self.expr(arg_type)
            self.match(RP)
            self.rpn += struct.pack("B", ptgParen)
        else:
            if (self.LA(1)==REF2D) and (_tokenSet_0.member(self.LA(2))):
                pass
                ref2d_tok = self.LT(1)
                self.match(REF2D)
                r, c = Utils.cell_to_packed_rowcol(ref2d_tok.text)
                if arg_type == "R":
                   self.rpn += struct.pack("<B2H", ptgRefR, r, c)
                else:
                   self.rpn += struct.pack("<B2H", ptgRefV, r, c)
            elif (self.LA(1)==REF2D) and (self.LA(2)==COLON):
                pass
                ref2d1_tok = self.LT(1)
                self.match(REF2D)
                self.match(COLON)
                ref2d2_tok = self.LT(1)
                self.match(REF2D)
                r1, c1 = Utils.cell_to_packed_rowcol(ref2d1_tok.text)
                r2, c2 = Utils.cell_to_packed_rowcol(ref2d2_tok.text)
                if arg_type == "R":
                   self.rpn += struct.pack("<B4H", ptgAreaR, r1, r2, c1, c2)
                else:
                   self.rpn += struct.pack("<B4H", ptgAreaV, r1, r2, c1, c2)
            elif (self.LA(1)==NAME) and (_tokenSet_0.member(self.LA(2))):
                pass
                name_tok = self.LT(1)
                self.match(NAME)
                self.rpn += ""
            elif (self.LA(1)==NAME) and (self.LA(2)==LP):
                pass
                func_tok = self.LT(1)
                self.match(NAME)
                if func_tok.text.upper() in std_func_by_name:
                   (opcode,
                   min_argc,
                   max_argc,
                   func_type,
                   arg_type_list,
                   volatile_func) = std_func_by_name[func_tok.text.upper()]
                else:
                   raise Exception, "unknown function: %s" % func_tok.text
                self.match(LP)
                arg_count=self.expr_list(arg_type_list, min_argc, max_argc)
                self.match(RP)
                if arg_count > max_argc or arg_count < min_argc:
                   raise Exception, "%d parameters for function: %s" % (arg_count, func_tok.text)
                if min_argc == max_argc:
                   if func_type == "V":
                       func_ptg = ptgFuncV
                   elif func_type == "R":
                       func_ptg = ptgFuncR
                   elif func_type == "A":
                       func_ptg = ptgFuncA
                   else:
                       raise Exception, "wrong function type"
                   self.rpn += struct.pack("<BH", func_ptg, opcode)
                else:
                   if func_type == "V":
                       func_ptg = ptgFuncVarV
                   elif func_type == "R":
                       func_ptg = ptgFuncVarR
                   elif func_type == "A":
                       func_ptg = ptgFuncVarA
                   else:
                       raise Exception, "wrong function type"
                   self.rpn += struct.pack("<2BH", func_ptg, arg_count, opcode)
            else:
                raise antlr.NoViableAltException(self.LT(1), self.getFilename())
            
    
    def expr_list(self,
        arg_type_list, min_argc, max_argc
    ):    
        arg_cnt = None
        
        arg_cnt = 0
        arg_type_list = arg_type_list.split()
        arg_type = arg_type_list[arg_cnt]
        la1 = self.LA(1)
        if False:
            pass
        elif la1 and la1 in [TRUE_CONST,FALSE_CONST,STR_CONST,NUM_CONST,INT_CONST,NAME,SUB,LP,REF2D]:
            pass
            self.expr(arg_type)
            arg_cnt += 1
            while True:
                if (self.LA(1)==SEMICOLON):
                    pass
                    if arg_cnt < len(arg_type_list):
                       arg_type = arg_type_list[arg_cnt]
                    else:
                       arg_type = arg_type_list[-1]
                    if arg_type == "...":
                       arg_type = arg_type_list[-2]
                    self.match(SEMICOLON)
                    la1 = self.LA(1)
                    if False:
                        pass
                    elif la1 and la1 in [TRUE_CONST,FALSE_CONST,STR_CONST,NUM_CONST,INT_CONST,NAME,SUB,LP,REF2D]:
                        pass
                        self.expr(arg_type)
                    elif la1 and la1 in [RP,SEMICOLON]:
                        pass
                        self.rpn += struct.pack("B", ptgMissArg)
                    else:
                            raise antlr.NoViableAltException(self.LT(1), self.getFilename())
                        
                    arg_cnt += 1
                else:
                    break
                
        elif la1 and la1 in [RP]:
            pass
        else:
                raise antlr.NoViableAltException(self.LT(1), self.getFilename())
            
        return arg_cnt
    

_tokenNames = [
    "<0>", 
    "EOF", 
    "<2>", 
    "NULL_TREE_LOOKAHEAD", 
    "TRUE_CONST", 
    "FALSE_CONST", 
    "STR_CONST", 
    "NUM_CONST", 
    "INT_CONST", 
    "NAME", 
    "EQ", 
    "NE", 
    "GT", 
    "LT", 
    "GE", 
    "LE", 
    "ADD", 
    "SUB", 
    "MUL", 
    "DIV", 
    "POWER", 
    "PERCENT", 
    "LP", 
    "RP", 
    "LB", 
    "RB", 
    "COLON", 
    "COMMA", 
    "SEMICOLON", 
    "CONCAT", 
    "REF2D"
]
    

### generate bit set
def mk_tokenSet_0(): 
    ### var1
    data = [ 817888258L, 0L]
    return data
_tokenSet_0 = antlr.BitSet(mk_tokenSet_0())
    



####################################################################################
#                                                                                  #
# Copyright (c) 2006 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#                                                                                  #
#  This file is globally imported into Picalo.  See picalo/__init__.py.            # 
#  When you write "from picalo import *", these functions all get imported.        #
#                                                                                  #
####################################################################################


import os, sys, time, types, codecs, gzip
import re, inspect, linecache, csv, xml.dom.minidom
from Table import Table
from TableArray import TableArray
import ZODB.DB
import persistent
from persistent.list import PersistentList


# global functions (defined in this file)
__all__ = [
  'TableList',
]
  

##########################################################
###   TableList type
###   A simple extension to list to allow lists of tables
###   to be identified without having to look through all
###   the elements to ensure they are Tables.

class TableList(persistent.Persistent):
  def __init__(self, *args, **kargs):
    '''A list of picalo tables.  Some functions in Picalo return a set
       of tables.  Sets of tables are returned as TableList objects.  
       
       Users do not normally create TableLists.  They are created automatically
       by Picalo functions.
       
       TableLists are a weaker form of TableArrays.  While TableArrays
       must have compatible tables, TableLists can hold tables of
       different schema.  TableLists can be viewed in the Picalo GUI,
       but they cannot be sent into most Picalo functions.
    '''
    self.changed = False
    if len(args) == 0:
      self._tables = PersistentList() 
    else:
      self._tables = PersistentList(args[0]) 
    for value in self:
      if not isinstance(value, (Table, TableArray, TableList)):
        raise TypeError, 'This type of list can only hold Picalo Tables, TableArrays, and TableLists.'
      
  
  def view(self):
    '''Opens the table list for viewing in the Picalo user interface.
       The resulting view allows you to page through the tables in the
       list.  See the first example.
       
       You can view individual tables in the list by using the [n]
       notation.  See the second example for this notation.
       
       Example:
        >>> data = Table([
        ...   ('id', int),
        ...   ('name', unicode),
        ... ],[
        ...   [ 1, 'Benny' ],
        ...   [ 2, 'Vijay' ],
        ... ])
        >>> tables = Grouping.stratify_by_value(data, 'id')
        >>> tables.view()
         
       Example:
        >>> data = Table([
        ...   ('id', int),
        ...   ('name', unicode),
        ... ],[
        ...   [ 1, 'Benny' ],
        ...   [ 2, 'Vijay' ],
        ... ])
        >>> tables = Grouping.stratify_by_value(data, 'id')
        >>> tables[0].view()
         
    '''
    from picalo import mainframe
    if mainframe != None:
      # try to figure out the variable name of this table by inspecting the call stack
      f = inspect.currentframe().f_back
      locals = f.f_locals
      for name in f.f_code.co_names:
        if locals.has_key(name) and locals[name] == self:
          mainframe.openTable(name)
          return
    # if we get here, either we're in console mode or we couldn't find the variable name
    print self
    
    
  def __repr__(self):
    '''For debugging.
       @return:   The number of tables in this TableList.
       @rtype:    str
    '''
    return '<TableList of %s: %s>' % (len(self), ', '.join([ repr(t) for t in self._tables]))
  
  def __str__(self):
    '''For debugging.
       @return:   The number of tables in this TableList.
       @rtype:    str
    '''
    return '<TableList of %s: %s>' % (len(self), ', '.join([ str(t) for t in self._tables]))
  
  
  ##### CONTAINER METHODS ####

  def __iter__(self):
    '''Returns an iterator to this TableList'''
    return TableListIterator(self)
        
        
  def __contains__(self, table):
    '''Returns whether the given table is in this TableArray'''
    if isinstance(table, (Table, TableArray)):
      for item in self:
        if item.get_columns() == table.get_columns() and item == table: # item == table checks the values of the tables
          return true
    return False
    
    
  def __eq__(self, other):
    '''Returns whether this TableArray is equal to a list/tuple/TableArray of tables.'''
    if other != None and len(self) == len(other):
      for i in range(len(self)):
        if self[i] != other[i]:
          return False
      return True
    return False
    

  def __ne__(self, other):
    '''Returns whether this TableArray is not equal to a list/tuple/TableArray of tables.'''
    return not self.__eq__(other)         
        
  
  def __len__(self):
    return len(self._tables)
    
  
  def __getitem__(self, index):
    return self._tables[index]
        

  def __setitem__(self, index, table):
    if not isinstance(table, Table):
      raise TypeError, 'This type of list can only hold Picalo tables.'
    self.changed = True
    self._tables[index] = table


  def __setslice__(self, i, j, sequence):
    for value in sequence:
      if not isinstance(value, Table):
        raise TypeError, 'This type of list can only hold Picalo tables.'
    self.changed = True
    self._table.__setslice__(i, j, sequence)


  def __getslice__(self, i, j):  
    return self._table.__getslice__(i, j)
    
    
  def append(self, table):
    if not isinstance(table, (Table, TableList, TableArray)):
      raise TypeError, 'This type of list can only hold Picalo tables.'
    self.changed = True
    self._tables.append(table)
    
    

  ###################################################################
  ###   Table functions that are applied to each table in the list

  def is_changed(self):
    '''Returns whether the table has been changed since loading'''
    if self.changed:
      return True
    for table in self:
      if table.is_changed():
        return True
    return False
    
    
  def set_changed(self, changed):
    '''Sets whether the class has been changed since loading.  This is not normally
       called by users.'''
    self.changed = True


  def set_readonly(self, readonly_flag=False):
    '''Sets the read only status of this table.  Tables that are read only cannot
       be modified.  Normally, tables are initially not read only (i.e. can be modified).
       The only exception is tables loaded from databases, which are read only.
       
       @param readonly_flag: True or False, depending upon whether the table should be read only or not.
       @type  readonly_flag: bool
    '''
    for table in self:
      table.set_readonly(readonly_flag)
  

  def is_readonly(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def filter(self, expression=None):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
      
      
  def clear_filter(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
       

  def is_filtered(self):
    '''Returns whether the first table in this list is filtered.'''
    return False


  def column(self, col):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
    

  def get_columns(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
    
    
  def get_column_names(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
    
    
  def column_count(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def set_name(self, column, name):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def set_type(self, column, column_type=None, format=None, expression=None):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def set_format(self, column, format=None):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def append_column(self, name, column_type, values=None):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
      
      
  def insert_column(self, index, name, column_type, values=None):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
      

  def append_calculated(self, name, expression):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
  
  
  def insert_calculated(self, index, name, expression):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
  
  
  def append_calculated_static(self, name, column_type, expression):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
  

  def insert_calculated_static(self, index, name, column_type, expression):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
  

  def move_column(self, column, new_index):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'


  def delete_column(self, column):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
  

  def structure(self):
    '''This method is not supported in TableLists.  Use TableArrays if this method is required.'''
    raise AssertionError, 'This method is not supported in TableLists.'
    
    
  def save(self, filename, respect_filter=False):
    '''Saves this TableList in native Picalo format.  This is the preferred
       format to save TableLists in because all column types, formulas, and so
       forth are saved.
    
       @param filename: The filename to save to.  This can also be an open stream.
       @type  filename: str
       @param respect_filter Whether to save the entire file or only those rows available through any current filters.
       @type  respect_filter bool
    '''
    # I can't import save until now because TableList class must be declared first
    from Table import save as table_save
    table_save(self, filename, respect_filter)


  def save_delimited(self, filename, delimiter=',', qualifier='"', line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in delimited format.  Since the table list probably has multiple tables
       in it, one delimited file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param delimiter:    An optional field delimiter character
       @type  delimiter:    str
       @param qualifier:    An optional qualifier to use when delimiters exist in field values
       @type  qualifier:    str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''    
    for i, table in enumerate(self):
      table.save_delimited('%05i-' % i + filename, delimiter, qualifier, line_ending, none, encoding, respect_filter)
      

  def save_csv(self, filename, line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in CSV format.  Since the table list probably has multiple tables
       in it, one CSV per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_csv('%05i-' % i + filename, line_ending, none, encoding, respect_filter)
    

  def save_tsv(self, filename, line_ending='\n', none='', encoding='utf-8', respect_filter=False):
    '''Saves this TableList in tsv format.  Since the table list probably has multiple tables
       in it, one tsv file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param encoding:     The unicode encoding to use for international or special characters.  For example, Microsoft applications like to use special characters for double quotes rather than the standard characters.  Unicode (the default) handles these nicely.
       @type  encoding:     str
    '''
    for i, table in enumerate(self):
      table.save_tsv('%05i-' % i + filename, line_ending, none, encoding, respect_filter)
    
    
  def save_fixed(self, filename, line_ending='\n', none='', respect_filter=False):
    '''Saves this TableList in fixed format.  Since the table list probably has multiple tables
       in it, one fixed file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with
       @type  line_ending:  str
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_fixed('%05i-' % i + filename, line_ending, none, respect_filter)
    
    
  def save_xml(self, filename, line_ending='\n', indent='\t', compact=False, none='', respect_filter=False):
    '''Saves this TableList in xml format.  Since the table list probably has multiple tables
       in it, one xml file per table is created by prepending 1, 2, 3 to the filename.

       @param filename:     A file name or a file pointer to an open file.  Using the file name string directly is suggested since it ensures the file is opened correctly for reading in CSV.
       @type  filename:     str
       @param line_ending:  An optional line ending to separate rows with (when compact is False)
       @type  line_ending:  str
       @param indent:       The character(s) to use for indenting (when compact is False)
       @type  indent:       str
       @param compact:      Whether to compact the XML or make it "pretty" with whitespace
       @type  compact:      bool
       @param none:         An optional parameter specifying what to write for cells that have the None value.
       @type  none:         str
       @param respect_filter Whether to save the entire file or only those rows available through any current filter.
       @type  respect_filter bool
    '''
    for i, table in enumerate(self):
      table.save_xml('%05i-' % i + filename, line_ending, indent, compact, none, respect_filter)
    
    
    
  
    
###############################
###   Iterator for a TableList
   
def TableListIterator(ta):     
  '''Returns a generator object to iterate over the tables in a TableList'''
  index = 0
  numcols = len(ta)
  while index < numcols:
    yield ta[index]
    index += 1  

    


####################################################################################
#                                                                                  #
# Copyright (c) 2007 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This software is free software; you can redistribute it and/or modify            #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# This software is distributed in the hope that it will be useful,                 #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################

__doc__ = '''
This module implements a memory-efficient CSV reader and writer.  Although 
Python 2.4+ has a built-in csv module, real-world tests by me show that it
is not memory efficient.  In other words, it runs out of memory on files
with 25 million or more records in it.

While this implementation may not be as fast (because it is written in
pure Python), it only holds one line at a time in memory.  It is up to
the calling application to make efficient use of these records.
'''

import os, cStringIO, codecs
from chardet.universaldetector import UniversalDetector


########################################################
###   Utility functions and variables

def make_unicode(value):
  '''Makes a value (of any type) into unicode using utf_8'''
  if isinstance(value, unicode):
    return value
  elif isinstance(value, str):
    return unicode(value, 'utf_8')
  else:
    return unicode(repr(value), 'utf_8')



########################################################
###   Delimited Writer

class DelimitedWriter:
  '''A writer of delimited text files'''
  def __init__(self, f, delimiter=',', qualifier='"', line_ending=os.linesep, none='', encoding='utf_8'):
    '''Constructor for a delimited writer object.
    
       @param f:            A file object that allows writing.  You must have already opened the file for writing.  This class does NOT close the writer.
       @type  f:            str
       @param delimiter:    A field delimiter character, defaults to a comma (,)
       @type  delimiter:    str
       @param qualifier:    A qualifier to use when delimiters exist in field records, defaults to a double quote (")
       @type  qualifier:    str
       @param line_ending:  A line ending to separate rows with, defaults to os.linesep (\n on Unix, \r\n on Windows)
       @type  line_ending:  str
       @param none:         An parameter specifying what to write for cells that have the None value, defaults to an empty string ('')
       @type  none:         str
       @param encoding:     The unicode encoding to write with.  This should be a value from the codecs module, defaults to 'utf_8'.
       @type  encoding:     str
    '''  
    self.f = codecs.EncodedFile(f, 'utf_8', encoding)
    self.delimiter = make_unicode(delimiter)
    self.qualifier = make_unicode(qualifier)
    self.doublequalifier = self.qualifier + self.qualifier
    self.line_ending = make_unicode(line_ending)
    self.none = none
    self.encoding = encoding


  def writerow(self, row, lastrow=False):
    '''Writes a row to the file.  This method is not thread-safe on the same DelimitedWriter object.
    
    @param row:     An iterable object to encode, such as a Python list or tuple.
    @type  row:     iterable object
    @param lastrow: Whether this is the last row of the file.  If True, no line ending is written.
    @type  lastrow: bool
    '''
    fields = []
    # encode for delimited text
    for val in row:
      val = make_unicode(val)
      if self.delimiter in val or self.qualifier in val or self.line_ending in val:
        fields.append(self.qualifier + val.replace(self.qualifier, self.doublequalifier) + self.qualifier)
      else:
        fields.append(val)
    
    # finally, write to the file in the desired encoding
    self.f.write(self.delimiter.join(fields).encode(self.encoding))
    if not lastrow:
      self.f.write(self.line_ending)
    self.f.flush()
    





###############################################
###   Delimited Reader

class DelimitedReader:
  '''A reader for delimited text files'''
  def __init__(self, f, delimiter=',', qualifier='"', encoding=None, errors='replace'):
    '''Constructor for a delimited text file reader.  Windows, Unix, and Mac line endings are automatically supported.
       The object can be iterated across (ex: for row in DelimitedReader(...) ) efficiently.  Each iteration produces
       a list of fields for the current row.  StopIteration is raised when no more records exist.

       @param f:            A file object that allows writing.  You must have already opened the file for writing.  This class does NOT close the writer.
       @type  f:            str
       @param delimiter:    A field delimiter character, defaults to a comma (,)
       @type  delimiter:    str
       @param qualifier:    A qualifier that denotes fields with special characters in them, defaults to a double quote (")
       @type  qualifier:    str
       @param encoding:     The unicode encoding to read with.  This should be a value from the codecs module.  If None, the encoding is guessed to utf_8, utf-16, utf-16-be, or utf-16-le
       @type  encoding:     str
       @param errors:       How to handle characters that cannot be decoded.  Options are 'replace', 'strict', and 'ignore'.  See 'codecs' in the Python documentation for more information.
       @type  errors:       str
    '''
    self.delimiter = delimiter
    self.qualifier = qualifier
    self.doublequalifier = qualifier + qualifier
    if encoding != None:
      self.encoding = encoding
    else: # we'll have to guess it with the chardet module
      startpos = f.tell()  # we'll have to read the file a bit, then reset at the end
      detector = UniversalDetector()
      while True:
        block = f.read(1024)  # read in 1K chunks
        detector.feed(block)
        if detector.done or not block:
          break
      detector.close()
      f.seek(startpos) # reset to the beginning position to read with the given encoding
      self.encoding = detector.result['encoding']
    # open the file with the given encoding
    self.f = codecs.EncodedFile(f, 'utf_8', self.encoding, errors)

    
  def __iter__(self):
    '''Returns an iterator to this DelimitedReader.  This allows code like the following:
    
       for record in delimitedtext.DelimitedReader(file):
         print record
    '''
    return self
    
    
  def _getnextline(self):
    '''Internal method that retrieves the next line of the file, separating the text from the line separator'''
    line = unicode(self.f.next(), 'utf_8')
    if line[-2:] in (u'\r\n', u'\n\r'):
      return line[:-2], line[-2:]
    elif line[-1:] in (u'\n', u'\r'):
      return line[:-1], line[-1:]
    return line, ''
    
  
  def _isendqualifier(self, field):
    '''Internal method that returns true if the given field ends with a qualifier.
       A field ends with a qualifier when it has an *odd* number
       of qualifiers at the end.  If it has an even number of qualifiers,
       the qualifiers are part of the actual field rather than a qualifier.
    '''
    qualifiers = self.qualifier
    count = 0
    while field.endswith(qualifiers):
      count += 1
      qualifiers += self.qualifier
    return count % 2 == 1
    

  def next(self):
    '''Reads a CSV record from the file and returns a list of fields encoded as unicode objects.
       A CSV record normally corresponds to a single line of the file, but it can span more than one
       line if the record contains hard returns in the values.  The method can handle these multi-line
       records.
       
       The method raises a StopIteration when done.  This supports the python Iterator interface.
       
       @returns  The fields from the next row.
       @rtype    list
    '''
    # The algorithm works by splitting the line by all delimiters first.  This will split some
    # (qualified) fields that shouldn't have been split.  It then goes through the splits 
    # and puts back together everything that shouldn't have been split.
    record = []
    combofield = None
    while True:
      line, linesep = self._getnextline()
      fields = line.split(self.delimiter)
      for field in fields:
        if combofield == None:    # we're not in quotes at the moment
          if field.startswith(self.qualifier) and self._isendqualifier(field[len(self.qualifier):]): # starts and ends with qualifier
            # remove the qualifiers and add now
            field = field[len(self.qualifier): -1*len(self.qualifier)]  # take off beginning and ending quotes
            record.append(field.replace(self.doublequalifier, self.qualifier))
        
          elif field.startswith(self.qualifier):  # starts with a qualifier but doesn't end yet
            # start quotes mode by assigning combofield (take off initial qualifier)
            combofield = field[len(self.qualifier):]
          
          else: # no qualifiers
            # this is just a regular field, so add it now
            record.append(field.replace(self.doublequalifier, self.qualifier))
        
        else:  # in quotes mode right now
          if self._isendqualifier(field[len(self.qualifier):]):  # we found the end qualifier, so add entire combofield now
            combofield += self.delimiter + field[:-1*len(self.qualifier)]  # take off ending quote
            record.append(combofield.replace(self.doublequalifier, self.qualifier))
            combofield = None
        
          else:  # not at end qualifier yet, so add to combofield and continue searching for end
            combofield += self.delimiter + field
            
      # if we're not in a combofield, we're done
      if combofield == None:
        return record
      
      # we're in a combofield, so add the line separator back into the file
      combofield += linesep


  def readrow(self):
    '''Reads a CSV record from the file and returns a list of fields encoded as unicode objects.
       A CSV record normally corresponds to a single line of the file, but it can span more than one
       line if the record contains hard returns in the values.  The method can handle these multi-line
       records.

       The method returns None when no more records exist.

       @returns  The fields from the next row.
       @rtype    list
    '''
    try:
      return self.next()
    except StopIteration:
      return None



#!/usr/bin/env python
# -*- coding: windows-1251 -*-

#  Copyright (C) 2005 Roman V. Kiseliov
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
# 
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
#  3. All advertising materials mentioning features or use of this
#     software must display the following acknowledgment:
#     "This product includes software developed by
#      Roman V. Kiseliov <roman@kiseliov.ru>."
# 
#  4. Redistributions of any form whatsoever must retain the following
#     acknowledgment:
#     "This product includes software developed by
#      Roman V. Kiseliov <roman@kiseliov.ru>."
# 
#  THIS SOFTWARE IS PROVIDED BY Roman V. Kiseliov ``AS IS'' AND ANY
#  EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL Roman V. Kiseliov OR
#  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
#  OF THE POSSIBILITY OF SUCH DAMAGE.


'''
From BIFF8 on, strings are always stored using UTF-16LE  text encoding. The
character  array  is  a  sequence  of  16-bit  values4.  Additionally it is
possible  to  use  a  compressed  format, which omits the high bytes of all
characters, if they are all zero.

The following tables describe the standard format of the entire string, but
in many records the strings differ from this format. This will be mentioned
separately. It is possible (but not required) to store Rich-Text formatting
information  and  Asian  phonetic information inside a Unicode string. This
results  in  four  different  ways  to  store a string. The character array
is not zero-terminated.

The  string  consists  of  the  character count (as usual an 8-bit value or
a  16-bit value), option flags, the character array and optional formatting
information.  If the string is empty, sometimes the option flags field will
not occur. This is mentioned at the respective place.

Offset  Size    Contents
0       1 or 2  Length of the string (character count, ln)
1 or 2  1       Option flags:
                  Bit   Mask Contents
                  0     01H  Character compression (ccompr):
                               0 = Compressed (8-bit characters)
                               1 = Uncompressed (16-bit characters)
                  2     04H  Asian phonetic settings (phonetic):
                               0 = Does not contain Asian phonetic settings
                               1 = Contains Asian phonetic settings
                  3     08H  Rich-Text settings (richtext):
                               0 = Does not contain Rich-Text settings
                               1 = Contains Rich-Text settings
[2 or 3] 2      (optional, only if richtext=1) Number of Rich-Text formatting runs (rt)
[var.]   4      (optional, only if phonetic=1) Size of Asian phonetic settings block (in bytes, sz)
var.     ln or 
         2·ln   Character array (8-bit characters or 16-bit characters, dependent on ccompr)
[var.]   4·rt   (optional, only if richtext=1) List of rt formatting runs 
[var.]   sz     (optional, only if phonetic=1) Asian Phonetic Settings Block 
'''


__rev_id__ = """$Id: UnicodeUtils.py,v 1.4 2005/07/20 07:24:11 rvk Exp $"""


import struct


DEFAULT_ENCODING = 'cp1251'

def u2ints(ustr):
    ints = [ord(uchr) for uchr in ustr]
    return ints

def u2bytes(ustr):
    ints = u2ints(ustr)
    return struct.pack('<' + 'H'*len(ints), *ints)

def upack2(_str):
    try:
        ustr = u2bytes(unicode(_str, 'ascii'))
        return struct.pack('<HB', len(_str), 0) + _str    
    except:
        if isinstance(_str, unicode):
            ustr = u2bytes(_str)
        else:
            ustr = u2bytes(unicode(_str, DEFAULT_ENCODING))
        return struct.pack('<HB', len(_str), 1) + ustr

def upack1(_str):
    try:
        ustr = u2bytes(unicode(_str, 'ascii'))
        return struct.pack('BB', len(_str), 0) + _str    
    except:
        if isinstance(_str, unicode):
            ustr = u2bytes(_str)
        else:
            ustr = u2bytes(unicode(_str, DEFAULT_ENCODING))
        return struct.pack('BB', len(_str), 1) + ustr

if __name__ == '__main__':   
    f = file('out0.bin', 'wb')
    f.write(u2bytes('þíèêîä: unicode'))
    f.close()

    f = file('out1.bin', 'wb')
    f.write(upack1('þíèêîä: unicode'))
    f.close()

    f = file('out2.bin', 'wb')
    f.write(upack2('þíèêîä: unicode'))
    f.close()






####################################################################################
#                                                                                  #
# Copyright (c) 2006 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################
#                                                                                  #
#  This file is globally imported into Picalo.  See picalo/__init__.py.            # 
#  When you write "from picalo import *", these functions all get imported.        #
#                                                                                  #
####################################################################################

from picalo import check_valid_table
from Error import error
from Table import Table
import types, os, os.path
import transaction
import ZODB.DB
import ZODB.FileStorage.FileStorage
import persistent
from persistent.dict import PersistentDict
from persistent.list import PersistentList

# global functions (defined in this file)
__all__ = [
  'Project',
  'create_project',
  'open_project',
  'save_project',
  'close_project',
  'delete_project',
]


####################################################################
###   Create, open, save, delete project.

def create_project(filename, cache_size=2000):
  '''Creates a new project, erasing any existing projects with 
     the filename.  Picalo projects normally have the .pcp
     extension.

     @param filename:   The filename the project will be stored in.
     @type  filename:   str
     @param cache_size: The number of objects (tables, records, columns, etc.) to keep in memory at any given time.
     @type  cache_size: int
     @return:           A Picalo project
     @rtype:            Project
  '''
  # erase any existing project
  if os.path.exists(filename):
    delete_project(filename)
  
  # create the new one
  return open_project(filename, cache_size)


def open_project(filename, cache_size=2000):
  '''Opens an existing project file in the workspace.
  
     @param filename:   The filename the project will be stored in.
     @type  filename:   str
     @param cache_size: The number of objects (tables, records, columns, etc.) to keep in memory at any given time.
     @type  cache_size: int
     @return:           A Picalo project
     @rtype:            Project
  '''
  return Project(filename, cache_size)
  

def save_project(project):
  '''Saves an open project.  This is a funcational alternative to
     project.save().'''
  project.save()
  
  
def close_project(project):
  '''Saves and closes a project, freeing any memory it is using.
     This is a funcational alternative to project.close().
  '''
  project.close()
  

def delete_project(filename):
  '''Removes the given project along with its associated indices and lock files.  The files
     and data within them are permanently removed from disk.'''
  assert isinstance(filename, types.StringTypes), 'Please enter a valid filename.'
  assert os.path.exists(filename), 'Project with filename %s does not exist.' % filename
  for ext in ( '', '.index', '.lock', '.tmp', '.old' ):
    if os.path.exists(filename + ext):
      os.remove(filename + ext)
  


####################################################################
###   Represents a Picalo project.  All data in Picalo are
###   kept in a project, and internally, it uses the Zope DB
###   to store everything.



class Project(persistent.Persistent):
  '''A top-level Picalo project.  All tables, database connections, queries,
     and other project settings are kept in a Project.  Picalo projects
     normally have the .pcp extension.
  '''
  def __init__(self, filename, cache_size=2000):
    '''Links to a project file on disk.  If the file does not exist, it 
       is created.
       
       @param filename:   The filename the project will be stored in.
       @type  filename:   str
       @param cache_size: The number of objects (tables, records, columns, etc.) to keep in memory at any given time.
       @type  cache_size: int
       @return:           A Picalo project
       @rtype:            Project
    '''
    assert filename, 'Please specify a filename for this project.'
    assert isinstance(filename, types.StringTypes), 'Please specify a valid filename for this project.'
    assert isinstance(cache_size, (int, long)), 'Please specify an integer for the cache size.'
    
    # open the database
    self.transaction_manager = transaction.TransactionManager()
    self.zodb = ZODB.DB(ZODB.FileStorage.FileStorage(filename), cache_size=cache_size)
    self.root = self.zodb.open(self.transaction_manager).root()
    
    # ensure we have the necessary elements for a Picalo project in the file
    if not self.root.has_key('Settings'):
      self.root['Settings'] = PersistentDict()
    if not self.root.has_key('Tables'):
      self.root['Tables'] = PersistentDict()
    if not self.root.has_key('DatabaseConnections'):
      self.root['DatabaseConnections'] = PersistentDict()
    if not self.root.has_key('Queries'):
      self.root['Queries'] = PersistentDict()
    
    
  def close(self):
    '''Closes the project file, automatically saving changes'''
    self.save()
    self.zodb.close()
    

  def save(self):
    '''Saves any changes to this project to disk.'''
    self.transaction_manager.commit()
    
    
  def abort(self):
    '''Restores the project to the state it was in when it was opened or just after the last save
       point, whichever is later.  All changes to the project, including its tables and information
       are lost.'''
    self.transaction_manager.abort()
    
    
  def pack(self):
    '''Packs the project to reduce its size on disk.  To make operations faster,
       Picalo files continue to grow over time, even when records and tables are 
       deleted.  Packing a project periodically ensures it doesn't become
       abnormally large.  Note that packing does not do any error correction in the
       file -- it simply removes any redundancies and deleted items.
       
       The following are done to pack a file:
        - Deleted records are permanently removed from their tables.
        - Deleted tables are permanently removed from the project.
    '''
    self.zodb.pack()
    
    
  def __str__(self):
    '''Prints a string representation of this project'''
    return '[Picalo Project: %s tables, %s database connections, %s queries]' % (len(self.root['Tables']), len(self.root['DatabaseConnections']), len(self.root['Queries']))
    
    
  def __getattr__(self, name):
    '''Retrieves a member of this project by name.  The order of search for the
       name is Project methods, tables, database connections, queries.
       
       Example:
         >>> # assume we have a project with a tabled named "mydata"
         >>> project.mydata.view()
         >>>
         >>> # alternatively:
         >>> project.get_table('mydata')
       
       @param col: The column name or index
       @type  col: str
       @return:    The value in the given field
       @rtype:     returns
    '''
    # class methods are already searched before __getattr__ gets called (not true with __setattr__, btw)
    for node in [ 'Tables', 'DatabaseConnections', 'Queries' ]:
      if self.root[node].has_key(name):
        return self.root[node][name]
    
    
      
    
  #########################################
  ###   Table methods
  
  def create_table(self, name, columns=3, data=[]):
    '''Creates a Picalo Table.  Tables are the single-most important item
       in Picalo.  It should be the first place users look to understand
       how to use the program.  See the manual for more information on
       how tables work.
       
       This method is the primary way to create tables in Picalo.  It is
       a convenience method that both creates the table and adds it to a 
       project.  
       
       See the Table() documentation for information on the format of the columns
       and data parameters.
    
       Example 1:
         >>> # create and assign a table to this project
         >>> data = project.create_table('data', ['id', 'name'])
         >>>
         >>> # the above line is equivalent to this:
         >>> data = Table(['id', 'name'])
         >>> project.add_table('data', data)
         >>>
         >>> # make the new table a permanent part of the project
         >>> project.save()

       @param name:      A name the table will be known by in this project.
       @type  name:      str
       @param columns:   A list of (name, type) pairs specifying the column names and their types
       @type  columns:   list
       @param data:      The initial data for the table specified as another Picalo table or a list of lists.
       @type  data:      Table/list
       @return:          The new Picalo table.
       @rtype:           Table
    '''
    assert isinstance(name, types.StringTypes) and name != '', 'Please enter a valid name for this table.'
    assert not self.root['Tables'].has_key(name), 'A table in this project already has the name "%s".  Please delete it first.' % name
    t = Table(columns, data)
    self.root['Tables'][name] = t
    return t
    
    
  def add_table(self, name, table):
    '''Adds an existing Picalo table to this project.  Tables that are created directly
       (with data = Table(...) syntax) are not automatically associated with a project.
       This method adds a table not associated with this project to this project.  When this
       project is saved, the table will be saved with it.
       
       Don't forget to call project.save() to make the change permanent.

       @param name:      A name the table will be known by in this project.
       @type  name:      str
       @return:          The Picalo table.
       @rtype:           Table
    '''
    assert isinstance(name, types.StringTypes) and name != '', 'Please enter a valid name for this table.'
    assert not self.root['Tables'].has_key(name), 'A table in this project already has the name "%s".  Please delete it first.' % name
    self.root['Tables'][name] = table


  def get_table(self, name):
    '''Returns a reference to a table with the given name.  Tables can also be referenced directly,
       as in project.tablename.

       @param name:      The name of the table within this project.
       @type  name:      str
       @return:          The Picalo table.
       @rtype:           Table
    '''
    assert isinstance(name, types.StringTypes), 'Please enter a valid name for this table.'
    assert self.root['Tables'].has_key(name), 'Table with name "%s" not found in the project.' % name
    return self.root['Tables'][name]
    
    
  def has_table(self, name):
    '''Returns whether this project has a table with the given name.
    
       @param name:      The name of the table within this project.
       @type  name:      str
       @return:          True if the table exists, False otherwise.
       @rtype:           boolean
    '''
    assert isinstance(name, types.StringTypes), 'Please enter a valid name for this table.'
    return self.root['Tables'].has_key(name)
    
    
  def list_tables(self):
    '''Returns an alphabetical list of tables in this project.

       @return:                   A list of table names.
       @rtype:                    list of str
    '''
    names = self.root['Tables'].keys()
    names.sort()
    return names
    
    
  def remove_table(self, name):
    '''Removes the table with the given name from the project.  
    
       Note that the table data are not removed from memory, so if you still have
       local variables pointing to the table, you still have full access to the 
       data.  This method just removes the table from the project file.

       Don't forget to call project.save() to make
       the change permanent.  Call project.pack() to free the space used by the 
       deleted table and make the project file smaller on disk.

       @param table:      The name of the table within this project.
       @type  table:      str
    '''
    assert isinstance(name, types.StringTypes), 'Please enter a valid name for this table.'
    assert self.root['Tables'].has_key(name), 'Table with name "%s" not found in the project.' % name
    del self.root['Tables'][name]
    
  


#!/usr/bin/env python
# -*- coding: windows-1251 -*-

#  Copyright (C) 2005 Roman V. Kiseliov
#  All rights reserved.
# 
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
# 
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
# 
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
# 
#  3. All advertising materials mentioning features or use of this
#     software must display the following acknowledgment:
#     "This product includes software developed by
#      Roman V. Kiseliov <roman@kiseliov.ru>."
# 
#  4. Redistributions of any form whatsoever must retain the following
#     acknowledgment:
#     "This product includes software developed by
#      Roman V. Kiseliov <roman@kiseliov.ru>."
# 
#  THIS SOFTWARE IS PROVIDED BY Roman V. Kiseliov ``AS IS'' AND ANY
#  EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#  PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL Roman V. Kiseliov OR
#  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
#  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
#  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
#  STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
#  OF THE POSSIBILITY OF SUCH DAMAGE.


'''
Record Order in BIFF8
  Workbook Globals Substream
      BOF Type = workbook globals
      Interface Header
      MMS
      Interface End
      WRITEACCESS
      CODEPAGE
      DSF
      TABID
      FNGROUPCOUNT
      Workbook Protection Block
            WINDOWPROTECT
            PROTECT
            PASSWORD
            PROT4REV
            PROT4REVPASS
      BACKUP
      HIDEOBJ 
      WINDOW1 
      DATEMODE 
      PRECISION
      REFRESHALL
      BOOKBOOL 
      FONT +
      FORMAT *
      XF +
      STYLE +
    ? PALETTE
      USESELFS
    
      BOUNDSHEET +
    
      COUNTRY 
    ? Link Table 
      SST 
      ExtSST
      EOF
'''


__rev_id__ = """$Id: Workbook.py,v 1.5 2005/10/26 07:44:24 rvk Exp $"""


import BIFFRecords
import Style
from Deco import accepts, returns


class Workbook(object):

    #################################################################
    ## Constructor
    #################################################################
    @accepts(object)
    def __init__(self):
        self.__owner = 'None'       
        self.__country_code = 0x07 
        self.__wnd_protect = 0
        self.__obj_protect = 0
        self.__protect = 0        
        self.__backup_on_save = 0
        # for WINDOW1 record
        self.__hpos_twips = 0x01E0
        self.__vpos_twips = 0x005A
        self.__width_twips = 0x3FCF
        self.__height_twips = 0x2A4E
        
        self.__active_sheet = 0
        self.__first_tab_index = 0
        self.__selected_tabs = 0x01
        self.__tab_width_twips = 0x0258
        
        self.__wnd_hidden = 0
        self.__wnd_mini = 0
        self.__hscroll_visible = 1
        self.__vscroll_visible = 1
        self.__tabs_visible = 1

        self.__styles = Style.StyleCollection()
         
        self.__dates_1904 = 0
        self.__use_cell_values = 1
        
        self.__sst = BIFFRecords.SharedStringTable()
        
        self.__worksheets = []

    #################################################################
    ## Properties, "getters", "setters"
    #################################################################

    @accepts(object, str)
    def set_owner(self, value):
        self.__owner = value

    def get_owner(self):
        return self.__owner

    owner = property(get_owner, set_owner)

    #################################################################

    @accepts(object, int)
    def set_country_code(self, value):
        self.__country_code = value

    def get_country_code(self):
        return self.__country_code

    country_code = property(get_country_code, set_country_code)

    #################################################################

    @accepts(object, bool)
    def set_wnd_protect(self, value):
        self.__wnd_protect = int(value)

    def get_wnd_protect(self):
        return bool(self.__wnd_protect)

    wnd_protect = property(get_wnd_protect, set_wnd_protect)

    #################################################################

    @accepts(object, bool)
    def set_obj_protect(self, value):
        self.__obj_protect = int(value)

    def get_obj_protect(self):
        return bool(self.__obj_protect)

    obj_protect = property(get_obj_protect, set_obj_protect)

    #################################################################

    @accepts(object, bool)
    def set_protect(self, value):
        self.__protect = int(value)

    def get_protect(self):
        return bool(self.__protect)

    protect = property(get_protect, set_protect)
    
    #################################################################

    @accepts(object, bool)
    def set_backup_on_save(self, value):
        self.__backup_on_save = int(value)

    def get_backup_on_save(self):
        return bool(self.__backup_on_save)

    backup_on_save = property(get_backup_on_save, set_backup_on_save)

    #################################################################

    @accepts(object, int)
    def set_hpos(self, value):
        self.__hpos_twips = value & 0xFFFF

    def get_hpos(self):
        return self.__hpos_twips

    hpos = property(get_hpos, set_hpos)

    #################################################################

    @accepts(object, int)
    def set_vpos(self, value):
        self.__vpos_twips = value & 0xFFFF

    def get_vpos(self):
        return self.__vpos_twips

    vpos = property(get_vpos, set_vpos)

    #################################################################

    @accepts(object, int)
    def set_width(self, value):
        self.__width_twips = value & 0xFFFF

    def get_width(self):
        return self.__width_twips

    width = property(get_width, set_width)

    #################################################################

    @accepts(object, int)
    def set_height(self, value):
        self.__height_twips = value & 0xFFFF

    def get_height(self):
        return self.__height_twips

    height = property(get_height, set_height)

    #################################################################

    @accepts(object, int)
    def set_active_sheet(self, value):
        self.__active_sheet = value & 0xFFFF
        self.__first_tab_index = self.__active_sheet

    def get_active_sheet(self):
        return self.__active_sheet

    active_sheet = property(get_active_sheet, set_active_sheet)

    #################################################################

    @accepts(object, int)
    def set_tab_width(self, value):
        self.__tab_width_twips = value & 0xFFFF

    def get_tab_width(self):
        return self.__tab_width_twips

    tab_width = property(get_tab_width, set_tab_width)

    #################################################################

    @accepts(object, bool)
    def set_wnd_visible(self, value):
        self.__wnd_hidden = int(not value)

    def get_wnd_visible(self):
        return not bool(self.__wnd_hidden)

    wnd_visible = property(get_wnd_visible, set_wnd_visible)

    #################################################################

    @accepts(object, bool)
    def set_wnd_mini(self, value):
        self.__wnd_mini = int(value)

    def get_wnd_mini(self):
        return bool(self.__wnd_mini)

    wnd_mini = property(get_wnd_mini, set_wnd_mini)

    #################################################################

    @accepts(object, bool)
    def set_hscroll_visible(self, value):
        self.__hscroll_visible = int(value)

    def get_hscroll_visible(self):
        return bool(self.__hscroll_visible)

    hscroll_visible = property(get_hscroll_visible, set_hscroll_visible)

    #################################################################

    @accepts(object, bool)
    def set_vscroll_visible(self, value):
        self.__vscroll_visible = int(value)

    def get_vscroll_visible(self):
        return bool(self.__vscroll_visible)

    vscroll_visible = property(get_vscroll_visible, set_vscroll_visible)

    #################################################################

    @accepts(object, bool)
    def set_tabs_visible(self, value):
        self.__tabs_visible = int(value)

    def get_tabs_visible(self):
        return bool(self.__tabs_visible)

    tabs_visible = property(get_tabs_visible, set_tabs_visible)

    #################################################################

    @accepts(object, bool)
    def set_dates_1904(self, value):
        self.__dates_1904 = int(value)

    def get_dates_1904(self):
        return bool(self.__dates_1904)

    dates_1904 = property(get_dates_1904, set_dates_1904)

    #################################################################

    @accepts(object, bool)
    def set_use_cell_values(self, value):
        self.__use_cell_values = int(value)

    def get_use_cell_values(self):
        return bool(self.__use_cell_values)

    use_cell_values = property(get_use_cell_values, set_use_cell_values)

    #################################################################

    def get_default_style(self):
        return self.__styles.default_style

    default_style = property(get_default_style)

    ##################################################################
    ## Methods
    ##################################################################

    @accepts(object, Style.XFStyle)
    def add_style(self, style):
        return self.__styles.add(style)

    @accepts(object, (str, unicode))    
    def add_str(self, s):
        return self.__sst.add_str(s)
        
    @accepts(object, str)    
    def str_index(self, s):
        return self.__sst.str_index(s)
        
    @accepts(object, (str, unicode))    
    def add_sheet(self, sheetname):
        import Worksheet
        self.__worksheets.append(Worksheet.Worksheet(sheetname, self))
        return self.__worksheets[-1]

    @accepts(object, int)    
    def get_sheet(self, sheetnum):
        return self.__worksheets[sheetnum]
        
    ##################################################################
    ## BIFF records generation
    ##################################################################

    def __bof_rec(self):
        return BIFFRecords.Biff8BOFRecord(BIFFRecords.Biff8BOFRecord.BOOK_GLOBAL).get()

    def __eof_rec(self):
        return BIFFRecords.EOFRecord().get()
        
    def __intf_hdr_rec(self):
        return BIFFRecords.InteraceHdrRecord().get()

    def __intf_end_rec(self):
        return BIFFRecords.InteraceEndRecord().get()

    def __intf_mms_rec(self):
        return BIFFRecords.MMSRecord().get()

    def __write_access_rec(self):
        return BIFFRecords.WriteAccessRecord(self.__owner).get()

    def __wnd_protect_rec(self):
        return BIFFRecords.WindowProtectRecord(self.__wnd_protect).get()

    def __obj_protect_rec(self):
        return BIFFRecords.ObjectProtectRecord(self.__obj_protect).get()

    def __protect_rec(self):
        return BIFFRecords.ProtectRecord(self.__protect).get()

    def __password_rec(self):
        return BIFFRecords.PasswordRecord().get()

    def __prot4rev_rec(self):
        return BIFFRecords.Prot4RevRecord().get()

    def __prot4rev_pass_rec(self):
        return BIFFRecords.Prot4RevPassRecord().get()

    def __backup_rec(self):
        return BIFFRecords.BackupRecord(self.__backup_on_save).get()
        
    def __hide_obj_rec(self):
        return BIFFRecords.HideObjRecord().get()
        
    def __window1_rec(self):
        flags = 0
        flags |= (self.__wnd_hidden) << 0
        flags |= (self.__wnd_mini) << 1
        flags |= (self.__hscroll_visible) << 3
        flags |= (self.__vscroll_visible) << 4
        flags |= (self.__tabs_visible) << 5
        
        return BIFFRecords.Window1Record(self.__hpos_twips, self.__vpos_twips, 
                                self.__width_twips, self.__height_twips, 
                                flags,
                                self.__active_sheet, self.__first_tab_index, 
                                self.__selected_tabs, self.__tab_width_twips).get()
        
    def __codepage_rec(self):
        return BIFFRecords.CodepageBiff8Record().get()
        
    def __country_rec(self):
        return BIFFRecords.CountryRecord(self.__country_code, self.__country_code).get()
        
    def __dsf_rec(self):
        return BIFFRecords.DSFRecord().get()
        
    def __tabid_rec(self):
        return BIFFRecords.TabIDRecord(len(self.__worksheets)).get()
        
    def __fngroupcount_rec(self):
        return BIFFRecords.FnGroupCountRecord().get()
        
    def __datemode_rec(self):
        return BIFFRecords.DateModeRecord(self.__dates_1904).get()        

    def __precision_rec(self):
        return BIFFRecords.PrecisionRecord(self.__use_cell_values).get()         

    def __refresh_all_rec(self):
        return BIFFRecords.RefreshAllRecord().get()        

    def __bookbool_rec(self):
        return BIFFRecords.BookBoolRecord().get()         

    def __all_fonts_num_formats_xf_styles_rec(self):
        return self.__styles.get_biff_data()

    def __palette_rec(self):
        result = ''
        return result
        
    def __useselfs_rec(self):
        return BIFFRecords.UseSelfsRecord().get()
        
    def __boundsheets_rec(self, data_len_before, data_len_after, sheet_biff_lens):
        #  .................................  
        # BOUNDSEHEET0
        # BOUNDSEHEET1
        # BOUNDSEHEET2
        # ..................................
        # WORKSHEET0
        # WORKSHEET1
        # WORKSHEET2
        boundsheets_len = 0
        for sheet in self.__worksheets:
            boundsheets_len += len(BIFFRecords.BoundSheetRecord(0x00L, sheet.hidden, sheet.name).get())
        
        start = data_len_before + boundsheets_len + data_len_after
        
        result = ''
        for sheet_biff_len,  sheet in zip(sheet_biff_lens, self.__worksheets):
            result += BIFFRecords.BoundSheetRecord(start, sheet.hidden, sheet.name).get()
            start += sheet_biff_len            
        return result

    def __all_links_rec(self):
        result = ''
        return result
        
    def __sst_rec(self):
        return self.__sst.get_biff_record()
        
    def __ext_sst_rec(self, abs_stream_pos):
        return ''
        #return BIFFRecords.ExtSSTRecord(abs_stream_pos, self.sst_record.str_placement,
        #self.sst_record.portions_len).get()

    def get_biff_data(self):
        before = ''
        before += self.__bof_rec()
        before += self.__intf_hdr_rec()
        before += self.__intf_mms_rec()
        before += self.__intf_end_rec()
        before += self.__write_access_rec()
        before += self.__codepage_rec()
        before += self.__dsf_rec() 
        before += self.__tabid_rec() 
        before += self.__fngroupcount_rec()
        before += self.__wnd_protect_rec()
        before += self.__protect_rec()
        before += self.__obj_protect_rec()
        before += self.__password_rec()
        before += self.__prot4rev_rec()
        before += self.__prot4rev_pass_rec()
        before += self.__backup_rec()        
        before += self.__hide_obj_rec()        
        before += self.__window1_rec()
        before += self.__datemode_rec()
        before += self.__precision_rec()
        before += self.__refresh_all_rec()
        before += self.__bookbool_rec()
        before += self.__all_fonts_num_formats_xf_styles_rec()
        before += self.__palette_rec()
        before += self.__useselfs_rec()
        
        country            = self.__country_rec()
        all_links          = self.__all_links_rec()
        
        shared_str_table   = self.__sst_rec()        
        after = country + all_links + shared_str_table
        
        ext_sst = self.__ext_sst_rec(0) # need fake cause we need calc stream pos
        eof = self.__eof_rec()

        self.__worksheets[self.__active_sheet].selected = True
        sheets = ''
        sheet_biff_lens = []
        for sheet in self.__worksheets:
            data = sheet.get_biff_data()
            sheets += data
            sheet_biff_lens.append(len(data))
            
        bundlesheets = self.__boundsheets_rec(len(before), len(after)+len(ext_sst)+len(eof), sheet_biff_lens)       
       
        sst_stream_pos = len(before) + len(bundlesheets) + len(country)  + len(all_links)
        ext_sst = self.__ext_sst_rec(sst_stream_pos)           
        
        return before + bundlesheets + after + ext_sst + eof + sheets

    def save(self, filename):
        import CompoundDoc

        doc = CompoundDoc.XlsDoc()
        doc.save(filename, self.get_biff_data())

if __name__ == '__main__':
    wb = Workbook()
    f = file('workbook.bin', 'wb')
    f.write(wb.get_biff_data())
    f.close()
    


#!/usr/bin/env python
# -*- coding: windows-1251 -*-

__rev_id__ = """$Id: __init__.py,v 1.6 2005/08/11 08:53:48 rvk Exp $"""

import sys
if sys.version_info[:2] < (2, 4):
    print >>sys.stderr, "Sorry, pyExcelerator requires Python 2.4 or later"
    sys.exit(1)


from Workbook import Workbook
from Worksheet import Worksheet
from Row import Row
from Column import Column 
from Formatting import Font, Alignment, Borders, Pattern, Protection
from Style import XFStyle 
from ImportXLS import * 

from ExcelFormula import *



#!/usr/bin/python

####################################################################################
#                                                                                  #
# Copyright (c) 2003 Dr. Conan C. Albrecht <conan_albrechtATbyuDOTedu>             #
#                                                                                  #
# This file is part of Picalo.                                                     #
#                                                                                  #
# Picalo is free software; you can redistribute it and/or modify                   #
# it under the terms of the GNU General Public License as published by             #
# the Free Software Foundation; either version 2 of the License, or                # 
# (at your option) any later version.                                              #
#                                                                                  #
# Picalo is distributed in the hope that it will be useful,                        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of                   #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                    #
# GNU General Public License for more details.                                     #
#                                                                                  #
# You should have received a copy of the GNU General Public License                #
# along with Foobar; if not, write to the Free Software                            #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA        #
#                                                                                  #
####################################################################################

# Note that the picalo/base/Global.py file modifies the dispatchers in
# stats.py to allow Picalo data types.


